import math

from scipy.special import comb

from pyloa.utils.math import s2
from .environment import PagingEnvironment


class DefaultPagingEnvironment(PagingEnvironment):
    """Default Paging Environment for pyloa.agent.PagingAgent agents to play on.

    Args:
        sequence_size (int): number of pages per sequence
        cache_size (int): number of pages in cache
        num_pages (int): number of different pages
        include_request (bool): depicts whether or not request (next page) is included within state
        include_pagefault (bool): depicts whether or not is_pagefault is included within state
        include_cache (bool): depicts whether or not _cache is included within state
        include_time_in_cache (bool): depicts whether or not _time_im_cache is included within state
        include_time_since_touched (bool): depicts whether or not _time_since_touched is included within state
        include_histogram (bool): depicts whether or not _histogram is included within state
    """
    def __init__(self,
                 sequence_size,
                 cache_size,
                 num_pages,
                 include_request=True,
                 include_pagefault=False,
                 include_cache=False,
                 include_time_in_cache=False,
                 include_time_since_touched=False,
                 include_histogram=False,
                 **kwargs):
        super().__init__(
            sequence_size=sequence_size,
            cache_size=cache_size,
            num_pages=num_pages,
            include_request=include_request,
            include_pagefault=include_pagefault,
            include_cache=include_cache,
            include_time_in_cache=include_time_in_cache,
            include_time_since_touched=include_time_since_touched,
            include_histogram=include_histogram,
            **kwargs
        )

    def _calculate_cache_state_complexity(self):
        """Calculates and returns the state complexity of cache states, if cache is included within environment's
        state representation; otherwise returns 1.

            #states = num_pages! * sum_{i=0}^{cache_size+1} binom{num_pages}{cache_size} / (num_pages - cache_size + i)!

        Returns:
            int: cardinality of cache-state space if included, else 1
        """
        if self._include_cache:
            _fac_num_pages = math.factorial(self.num_pages)
            res = 0
            for i in range(0, self._cache_size+1):
                _comb = comb(N=self.num_pages, k=self._cache_size, exact=True)
                # Todo: is this correct? is _comb really independent from i??? -> move it out of the loop at least?
                res += (_fac_num_pages * _comb) // math.factorial(self.num_pages - self._cache_size + i)
        else:
            res = 1
        return res

    def _calculate_time_in_cache_state_complexity(self):
        """Calculates and returns the state complexity of time_in_cache states, if time_in_cache is included within
        environment's state representation; otherwise returns 1.

            #states = _sequence_size ** _cache_size

        Returns:
            int: cardinality of time_in_cache-state space if included, else 1
        """
        if self._include_time_in_cache:
            res = self._sequence_size ** self._cache_size
            # Todo: this is not correct (time_in_cache has unique values)
        else:
            res = 1
        return res

    def _calculate_time_since_touched_state_complexity(self):
        """Calculates and returns the state complexity of time_since_touched states, if time_since_touched is included
        within environment's state representation; otherwise returns 1.

            # states = sum_{c=1}^{cache_size+1} binom{cache_size}{c} * sum_{k=1}^{sequence_size+1} S(k, c)

        whereas S denotes the Stirling number of the second kind (pyloa.utils.s2).

        Returns:
            int: cardinality of time_since_touched-state space if included, else 1
        """
        if self._include_time_since_touched:
            res = 0
            for c in range(1, self._cache_size+1):
                _comb = comb(N=self._cache_size, k=c, exact=True)
                _sum = 0
                for k in range(1, self._sequence_size+1):
                    _sum += s2(k, c)
                res += _sum * _comb
        else:
            res = 1
        return res

    def _calculate_histogram_state_complexity(self):
        """Calculates and returns the state complexity of histogram states, if histogram is included within
        environment's state representation; otherwise returns 1.

            # states = sum_{c=1}^{cache_size+1} binom{cache_size}{c} * sum_{k=1}^{sequence_size+1} S(k, c)

        whereas S denotes the Stirling number of the second kind (pyloa.utils.s2).

        Returns:
            int: cardinality of time_since_touched-state space if included, else 1
        """
        if self._include_histogram:
            res = 0
            for c in range(1, self._cache_size+1):
                _comb = comb(N=self._cache_size, k=c, exact=True)
                _sum = 0
                for k in range(1, self._sequence_size+1):
                    _sum += s2(k, c)
                res += _sum * _comb
        else:
            res = 1
        return res



from abc import abstractmethod

import numpy as np

from .agent import Agent
from .rlagent import RLAgent
from .qagent import QAgent
from pyloa.utils.tensorboard import tensorboard_array, tensorboard_scalar, tensorboard_histo


class VertexAgent(Agent):
    """Abstract base class for Agents that play on pyloa.environment.VertexEnvironment instances.

    Args:
        state_dim (int): defines input shape=(_state_dim,) of input states
        action_dim (int): defines the number of different actions the agent can take
    """
    def __init__(self, state_dim, action_dim, **kwargs):
        super().__init__(state_dim=state_dim, action_dim=action_dim)

    # noinspection PyUnresolvedReferences
    # noinspection PyAttributeOutsideInit
    def play_instance(self, environment, is_train, episode, writer, invert_cost=True):
        """Agent plays on environment for one episode (on one graph).

        Args:
            environment (pyloa.environment.VertexEnvironment): environment to play on
            episode (int): number depicting how many episodes agent played so far
            is_train (bool): depicts if agent should train (True) or just evaluate (False)
            writer (tf.summary.FileWriter): tensorboard logging for playing an instance
            invert_cost (bool): depicts if costs are inverted after calculation (cost = negative reward)

        Returns:
            tuple: agent will return its calculated metrics of this episode as follows (steps, cost, page_faults)
                    steps (int): depicts the number of actions till the environment terminated
                    cost (int): total cost accumulated during episode
                    page_faults (int): depicts the number of page faults the agent had during episode

        Raises:
            RuntimeError: if agent is invoked with is_train=True, but agent is not a subclass of RLAgent
        """
        _is_trainable = isinstance(self, RLAgent)         # has epsilon and train
        _has_memory = isinstance(self, QAgent)            # has epsilon, memory and train
        # temp vars
        _temp_cost = 0
        _temp_actions = []
        _temp_losses = []
        _temp_epsilon = 0.0
        _temp_local_cognition_fault = 0

        # sanity check to make sure only trainable agents can be trained
        if is_train and not _is_trainable:
            raise RuntimeError("{} called to learn but agent isn't trainable!".format(self.__class__.__name__))

        # change epsilon if agent is trainable but isn't training right now for eval
        if _is_trainable and not is_train:
            _temp_epsilon = self.epsilon
            self.epsilon = 1.0

        # play environment till it terminates
        while not environment.is_terminated:
            # get current state
            state = environment.get_state(vectorized_state=_is_trainable)

            # get next predicted action
            action = self.act(state=state)

            # take action
            next_state, cost = environment.step(action=action, vectorized_state=_is_trainable)
            cost = -cost if invert_cost else cost

            # update action_histogram, cost temp metrics, time in cache
            _temp_local_cognition_fault += (not environment.was_locally_colorized)
            _temp_actions.append(action)
            _temp_cost += abs(cost)

            # store action if agent is training
            if _has_memory and is_train:
                self.store(state=state, action=action, reward=cost, next_state=next_state)

            # train episodic non-QAgents, append loss values
            if not _has_memory and _is_trainable and is_train:
                _loss = self.train(state=state, action=action, reward=cost, next_state=next_state)
                _temp_losses.append(_loss)

        # train replay memory QAgents
        if _has_memory and is_train:
            _temp_losses = self.train(train_writer=writer)

        # adapt epsilon
        if _is_trainable:
            if is_train:
                self.increment_epsilon()        # increment epsilon if agent is trainable and is training
            else:
                self.epsilon = _temp_epsilon    # change epsilon back if agent is trainable but isn't training right now

        # store metrics
        if writer is not None:
            tensorboard_scalar(writer, 'total_cost/' + str(self), _temp_cost, episode)
            tensorboard_scalar(writer, 'avg_cost/' + str(self), _temp_cost / environment.num_steps, episode)
            tensorboard_scalar(writer, 'local_cognition_fault_ratio/' + str(self), _temp_local_cognition_fault /
                               environment.num_steps, episode)
            tensorboard_scalar(writer, 'num_colors/' + str(self), environment.num_colors, episode)
            tensorboard_scalar(writer, 'global_colorizing/' + str(self), environment.is_colorized, episode)
            tensorboard_histo(writer, 'actions/' + str(self), _temp_actions, episode)
            if is_train:
                tensorboard_array(writer, 'episode/loss', _temp_losses, episode)
                tensorboard_scalar(writer, 'episode/epsilon', self.epsilon, episode)

        # return performance
        return environment.num_steps, _temp_cost, 0

    @abstractmethod
    def act(self, state):
        """Agent decides on an action based on information given with state.

        Args:
            state (np.ndarray with shape=(_state_dim,)): input state

        Returns:
            int: action to be taken

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")


class FIFOVertexAgent(VertexAgent):
    """Deterministic First-In-First-Out VertexAgent.

    Args:
        state_dim (int): defines input shape=(_state_dim,) of input states
        action_dim (int): defines the number of different actions the agent can take
    """
    def __init__(self, state_dim, action_dim, **kwargs):
        super().__init__(state_dim=state_dim, action_dim=action_dim, **kwargs)

    def act(self, state):
        """Agent uses smallest-numerical color, which is unused in neighbourhood of order 1.

        Args:
            state (dict): input state

        Returns:
            int: action to be taken

        Raises:
            AssertionError: if current graph is not colorable by FIFOVertexAgent
        """
        _all_colors = list(range(1, 1+len(state['color_flag'])))        # create all colors
        _neighbour_colors = set(state['delta'][state['delta'] > 0])     # get neighbourhood colors
        for n in _neighbour_colors:                                     # remove neighbourhood colors
            _all_colors.remove(n)
        assert len(_all_colors) > 0, "{} can not color current graph"
        return _all_colors[0] - 1                                       # return lowest color


class RandomVertexAgent(VertexAgent):
    """Non-Deterministic RandomVertexAgent, which picks a random color each time.

    Args:
        state_dim (int): defines input shape=(_state_dim,) of input states
        action_dim (int): defines the number of different actions the agent can take
    """
    def __init__(self, state_dim, action_dim, seed=None, **kwargs):
        self._seed = seed
        super().__init__(state_dim=state_dim, action_dim=action_dim, **kwargs)
        self._rnd_state = np.random.RandomState(seed=seed)

    def act(self, state):
        """Agent picks a random available color.

        Args:
            state (dict): input state

        Returns:
            int: action to be taken
        """
        return self._rnd_state.randint(low=0, high=len(state['color_flag']))

    def __str__(self):
        """str: string representation of self"""
        return self.__class__.__name__ + "(seed={})".format(self._seed)


import math

import numpy as np
from scipy.stats import rankdata
from scipy.special import comb

from .environment import PagingEnvironment


class NormalizedPagingEnvironment(PagingEnvironment):
    """Normalized Paging Environment for pyloa.agent.PagingAgent agents to play on.

    Normalized Paging Environment ranks its internal structures (_histogram, _time_in_cache, _time_since_touched) when
    they are being requested via their getters. _rank_method depicts how ranks are assigned to equal values ('average',
    'min', 'max', 'dense', 'ordinal').

    Args:
        sequence_size (int): number of pages per sequence
        cache_size (int): number of pages in cache
        num_pages (int): number of different pages
        include_request (bool): depicts whether or not request (next page) is included within state
        include_pagefault (bool): depicts whether or not is_pagefault is included within state
        include_cache (bool): depicts whether or not _cache is included within state
        include_time_in_cache (bool): depicts whether or not _time_im_cache is included within state
        include_time_since_touched (bool): depicts whether or not _time_since_touched is included within state
        include_histogram (bool): depicts whether or not _histogram is included within state

    Attributes:
        _rank_method (str): depicts how ranks are assigned to equal values, check scipy.stats.rankdata for reference
        _cache_time_ranks (np.ndarray, shape=(_cache_size,), dtype=np.int32): ranking of _time_in_cache
        _touched_time_ranks (np.ndarray, shape=(_cache_size,), dtype=np.int32): ranking of _time_since_touched
        _histogram_ranks (np.ndarray, shape=(_cache_size,), dtype=np.int32): ranking of _histogram
    """
    def __init__(self,
                 sequence_size,
                 cache_size,
                 num_pages,
                 include_request=False,
                 include_pagefault=False,
                 include_cache=False,
                 include_time_in_cache=False,
                 include_time_since_touched=False,
                 include_histogram=False,
                 rank_method='dense',
                 **kwargs):
        super().__init__(
            sequence_size=sequence_size,
            cache_size=cache_size,
            num_pages=num_pages,
            include_request=include_request,
            include_pagefault=include_pagefault,
            include_cache=include_cache,
            include_time_in_cache=include_time_in_cache,
            include_time_since_touched=include_time_since_touched,
            include_histogram=include_histogram,
            **kwargs
        )
        # params
        self._rank_method = rank_method
        # structures
        self._cache_time_ranks = np.zeros(self._cache_size, dtype=np.int32)
        self._touched_time_ranks = np.zeros(self._cache_size, dtype=np.int32)
        self._histogram_ranks = np.zeros(self._cache_size, dtype=np.int32)

    def _calculate_cache_state_complexity(self):
        """Calculates and returns the state complexity of cache states, if cache is included within environment's
        state representation; otherwise returns 1.

            #states = num_pages! * sum_{i=0}^{cache_size+1} binom{num_pages}{cache_size} / (num_pages - cache_size + i)!

        Returns:
            int: cardinality of cache-state space if included, else 1
        """
        if self._include_cache:
            _fac_num_pages = math.factorial(self.num_pages)
            res = 0
            for i in range(0, self._cache_size+1):
                _comb = comb(N=self.num_pages, k=self._cache_size, exact=True)
                # Todo: is this correct? is _comb really independent from i??? -> move it out of the loop at least?
                res += (_fac_num_pages * _comb) // math.factorial(self.num_pages - self._cache_size + i)
        else:
            res = 1
        return res

    # Todo: define ranked time complexity
    def _calculate_time_in_cache_state_complexity(self):
        raise NotImplementedError()

    # Todo: define ranked time since touched complexity
    def _calculate_time_since_touched_state_complexity(self):
        raise NotImplementedError()

    # Todo: define ranked histogram complexity
    def _calculate_histogram_state_complexity(self):
        raise NotImplementedError()

    def _get_time_in_cache(self, maximum_first=True):
        """Calculates and returns the rank of values in _time_in_cache. Placeholder values in _cache have rank 0.

        Args:
            maximum_first (bool): if True, highest value will have rank 1; else highest value will have last rank.

        Returns:
            np.ndarray: rank for each index in _time_in_cache
        """
        return self._rank_array(self._time_in_cache, self._cache_time_ranks, maximum_first)

    def _get_time_since_touched(self, maximum_first=True):
        """Calculates and returns the rank of values in _time_since_touched. Placeholder values in _cache have rank 0.

        Args:
            maximum_first (bool): if True, highest value will have rank 1; else highest value will have last rank.

        Returns:
            np.ndarray: rank for each index in _time_since_touched
        """
        return self._rank_array(self._time_since_touched, self._touched_time_ranks, maximum_first)

    def _get_histogram(self, maximum_first=True):
        """Calculates and returns the rank of values in _histogram. Placeholder values in _cache have rank 0.

        Args:
            maximum_first (bool): if True, highest value will have rank 1; else highest value will have last rank.

        Returns:
            np.ndarray: rank for each index in _histogram
        """
        return self._rank_array(self._histogram, self._histogram_ranks, maximum_first)

    def _rank_array(self, source_array, target_array, maximum_first=True):
        """Calculates the rank of values in source_array and places their rank in target_array. Placeholder values in
        _cache have rank 0.

        Args:
            maximum_first (bool): if True, highest value will have rank 1; else highest value will have last rank.

        Returns:
            np.ndarray: rank for each index in source_array
        """
        target_array.fill(0)
        _used_indices = self.used_indices
        _used_source = source_array[_used_indices]
        _used_source = np.iinfo(np.int32).max - _used_source if maximum_first else _used_source
        _ranked_used_source = rankdata(_used_source, method=self._rank_method)
        target_array[_used_indices] = _ranked_used_source
        return target_array

    time_in_cache = property(fset=None, fget=_get_time_in_cache, fdel=None)
    time_since_touched = property(fset=None, fget=_get_time_since_touched, fdel=None)
    histogram = property(fset=None, fget=_get_histogram, fdel=None)



import numpy as np

from .environment import Environment
from pyloa.utils.graph import Graph


class VertexEnvironment(Environment):
    """Base class for pyloa.agent.VertexAgent agents to play on.

    Args:
        max_delta (int): maximum degree in graph that a vertex can have
        max_color (int): number of colors that graph is allowed to be colored with
        include_vertex (bool): depicts whether or not the current vertex is included within state
        include_delta (bool): depicts whether or not first order neighborhood colors are included within state
        include_delta_squared (bool): depicts whether or not second order neighborhood colors are included within state
        include_color_flag (bool): depicts whether or not _color flags is included within state
        include_num_steps (bool): depicts whether or not _num_steps is included within state
        include_embedding (bool): depicts whether or not embedding is included within state
        include_vertex_coloring (bool): depicts whether or not current graph coloring is included within state
        delta_padding_value (int): default value for delta and delta² if vertex has less than max_delta neighbors

    Attributes:
        _max_delta (int): maximum degree in graph that a vertiex can have
        _max_color (int): number of colors that graph is allowed to be colored with
        _include_vertex (bool): depicts whether or not the current vertex is included within state
        _include_delta (bool): depicts whether or not first order neighborhood colors are included within state
        _include_delta_squared (bool): depicts whether or not second order neighborhood colors are included within state
        _include_color_flag (bool): depicts whether or not _color flags is included within state
        _include_num_steps (bool): depicts whether or not _num_steps is included within state
        _include_embedding (bool): depicts whether or not embedding is included within state
        _include_vertex_coloring (bool): depicts whether or not current graph coloring is included within state
        _delta_padding_value (int): default value for delta and delta² if vertex has less than max_delta neighbors
        _last_state (dict or np.ndarray): the previous state of self
        _graph (pyloa.utils.Graph): the graph instance agents plas on (tries to colorize)
        _vertex_order (list of int): the order, in which environment will iterate over graph
        _num_steps (int): number of steps already played on current graph instance
        _color_flags (np.npdarray, shape=(_max_color,), dtype=np.int32): depicts if color at index has been used yet

    Raises:
        NotImplementedError: if abstract class is to be instanced
    """
    def __init__(self,
                 max_delta,
                 max_color,
                 include_vertex=True,
                 include_delta=True,
                 include_delta_squared=False,
                 include_color_flag=True,
                 include_num_steps=False,
                 include_embedding=False,
                 include_vertex_coloring=False,
                 delta_padding_value=-1,
                 **kwargs):
        super(VertexEnvironment, self).__init__()
        assert max_color >= max_delta
        # params
        self._max_delta = max_delta
        self._max_color = max_color
        self._delta_padding_value = delta_padding_value
        # state config
        self._include_vertex = include_vertex
        self._include_delta = include_delta
        self._include_delta_squared = include_delta_squared
        self._include_color_flag = include_color_flag
        self._include_num_steps = include_num_steps
        self._include_embedding = include_embedding
        self._include_vertex_coloring = include_vertex_coloring
        # structures
        self._last_state = None
        self._graph = None
        self._vertex_order = list(range(max_delta))
        self._num_steps = 0
        self._color_flags = np.zeros(self._max_color, dtype=np.int32)

    def _get_embedding(self):
        """np.ndarray: global embedding of the graph"""
        raise NotImplementedError("not implemented yet")

    def _is_terminated(self):
        """bool: True, if environment is terminated (all requests for _sequence evaluated). False otherwise."""
        return self._num_steps > 0 and self._vertex_order is not None and self._num_steps == len(self._vertex_order)

    def _is_colorized(self):
        """bool: True, if graph is correctly colorized. False otherwise."""
        return self._graph.is_colorized()

    def _is_locally_colorized(self):
        """bool: True, if graph at self.vertex is correctly locally colorized. False otherwise."""
        return self._graph.is_locally_colorized(self.vertex)

    def _was_locally_colorized(self):
        """bool: True, if graph at self.last_vertex is correctly locally colorized. False otherwise."""
        return self._graph.is_locally_colorized(self.last_vertex)

    def _get_action_dim(self):
        """int: number of actions"""
        return self._max_color

    def _get_state_dim(self):
        """int: length of state vector"""
        _sum = 0
        if self._include_vertex:
            _sum += 1
        if self._include_num_steps:
            _sum += 1
        if self._include_delta:
            _sum += self._max_delta
        if self._include_delta_squared:
            _sum += self._max_delta ** 2
        if self._include_color_flag:
            _sum += len(self._color_flags)
        if self._include_embedding:
            raise NotImplementedError("no embedding implementation yet in graph")
        if self._include_vertex_coloring and self._graph is not None:
            _sum += len(self._graph.coloring)
        return _sum

    def _get_delta(self):
        """np.ndarray: returns the first order neighborhood colors of self.vertex"""
        _neighbour_colors = self._graph.get_neighbours_colors(self.vertex)
        _res = np.ones(self._max_delta, dtype=np.int32) * self._delta_padding_value
        _res[0:len(_neighbour_colors)] = _neighbour_colors
        return _res

    def _get_delta_squared(self):
        """np.ndarray: returns the second order neighborhood colors of self.vertex"""
        _neighbours = self._graph[self.vertex]
        _res = np.empty(0, dtype=np.int32)
        for idx in range(self._max_delta):
            _tmp = np.ones(self._max_delta, dtype=np.int32) * self._delta_padding_value
            if idx < len(_neighbours):
                _neighbour_colors = self._graph.get_neighbours_colors(_neighbours[idx])
                _tmp[0:len(_neighbour_colors)] = _neighbour_colors
            _res = np.append(_res, _tmp)
        return _res

    def _get_instance(self):
        """Graph: returns the currently played on graph instance"""
        return self._graph

    def _get_num_colors(self):
        """int: returns the number of currently used colors on current graph instance"""
        return np.sum(self._color_flags)

    def _get_num_steps(self):
        """int: returns the number of played steps on current graph instance"""
        return self._num_steps

    def _get_current_vertex(self):
        """int: returns the current vertex to colotize for the agent"""
        return self._vertex_order[min(self._num_steps, len(self._vertex_order)-1)]

    def _get_last_vertex(self):
        """int: returns the previous vertex that has been colored with the last step"""
        if self.is_terminated:
            return self._vertex_order[-1]
        return self._vertex_order[min(self._num_steps, len(self._vertex_order)-1)-1]

    def _set_instance(self, value):
        """Overrides currently stored _graph with value without invoking reset()

        Args:
            value (pyloa.utils.Graph): new graph instance to play on

        Raises:
            AssertionError: if value is not a Graph instance
        """
        assert isinstance(value, Graph)
        self._graph = value.copy()
        self._vertex_order = [v for v in self._graph]
        self._num_steps = 0
        self._color_flags[np.flatnonzero(self._graph.coloring)] = 1

    def get_state(self, vectorized_state=False, vectorized_dtype=np.int32):
        """Returns the state of environment. The state is either vectorized (returns np.ndarray) or not (returns dict).
        The state comprises copies of each variable that has been defined to be included during environment creation.

        Args:
            vectorized_state: defines return type. A vectorized_state will be of type np.ndarray, otherwise dict
            vectorized_dtype: dtype of np.darray for vectorized_state

        Returns:
            np.ndarray or dict: state depicting the current observation (changing the returned state won't change the
                internal structures in environment)
        """
        if self.is_terminated:
            if vectorized_state:
                return np.ones_like(self._last_state) * (-2)
            else:
                return dict()

        if vectorized_state:
            _state = np.empty(0, dtype=vectorized_dtype)
            if self._include_vertex:
                _state = np.append(_state, np.asarray(self.vertex, dtype=vectorized_dtype))
            if self._include_num_steps:
                _state = np.append(_state, np.asarray(self._num_steps, dtype=vectorized_dtype))
            if self._include_delta:
                _state = np.append(_state, self.delta)
            if self._include_delta_squared:
                _state = np.append(_state, self.delta_squared)
            if self._include_color_flag:
                _state = np.append(_state, self._color_flags)
            if self._include_embedding:
                _state = np.append(_state, np.asarray(self.embedding, dtype=vectorized_dtype))
            if self._include_vertex_coloring:
                _state = np.append(_state, self._graph.coloring)
        else:
            _state = dict()
            _state['current_vertex'] = self.vertex
            _state['num_steps'] = self._num_steps
            _state['delta'] = self.delta
            _state['delta_squared'] = self.delta_squared
            _state['color_flag'] = np.copy(self._color_flags)
            _state['embedding'] = self._graph
            _state['vertex_coloring'] = np.copy(self._graph.coloring)
        self._last_state = _state
        return _state

    def calculate_state_complexity(self):
        raise NotImplementedError("not implemented yet")

    def step(self, action, vectorized_state=True):
        """Environment computes the step for action, updates all internal structures and returns the next state and cost
        as a tuple.

        Args:
            action (int): depicts the action (color) that the agent wants color self.vertex with
            vectorized_state (bool): depicts the return type of the observation after step/action is taken

        Returns:
            tuple: first entry is the next state, second entry are the costs affiliated with last action in last state

        Raises:
            AssertionError: if action is not in interval [0, max_color-1] and therefore invalid
        """
        assert 0 <= action < self._max_color, "action not in bounds: {}".format(action)
        picked_color = action + 1
        neighbour_colors = self._graph.get_neighbours_colors(self.vertex)
        if picked_color in neighbour_colors:
            cost = 1000
        elif self._color_flags[action] == 0:
            cost = 250
        else:
            cost = 0
        self._graph.coloring[self.vertex] = picked_color
        self._color_flags[action] = 1
        self._num_steps += 1
        return self.get_state(vectorized_state=vectorized_state), cost

    def reset(self):
        """Resets the environment to its initial state."""
        self._graph = None
        self._last_state = None
        self._num_steps = 0
        self._vertex_order.clear()
        self._color_flags[:] = 0

    vertex = property(fset=None, fget=_get_current_vertex, fdel=None)
    last_vertex = property(fset=None, fget=_get_last_vertex, fdel=None)
    embedding = property(fset=None, fget=_get_embedding, fdel=None)
    delta = property(fset=None, fget=_get_delta, fdel=None)
    delta_squared = property(fset=None, fget=_get_delta_squared, fdel=None)
    instance = property(fset=_set_instance, fget=_get_instance, fdel=None)
    is_terminated = property(fset=None, fget=_is_terminated, fdel=None)
    is_locally_colorized = property(fset=None, fget=_is_locally_colorized, fdel=None)
    was_locally_colorized = property(fset=None, fget=_was_locally_colorized, fdel=None)
    is_colorized = property(fset=None, fget=_is_colorized, fdel=None)
    num_colors = property(fset=None, fget=_get_num_colors, fdel=None)
    action_dim = property(fset=None, fget=_get_action_dim, fdel=None)
    state_dim = property(fset=None, fget=_get_state_dim, fdel=None)
    num_steps = property(fset=None, fget=_get_num_steps, fdel=None)



import os
import logging
from shutil import copyfile
from collections import OrderedDict
from pathlib import Path

import tensorflow as tf
from tqdm import tqdm

from pyloa.utils import str_iterable, tensorboard_text, load_hyperparameters


def run(args, params, dirs):
    """Sets up all necessary objects and containers before training.

    Firstly, instantiates the necessary objects defined in params (agent, instance_generator, environment); secondly.
    backs up the configuration file and starts to eval all agents. Agents are loaded both from pre-trained agents (all
    hyperparams definitons nested in root_dir) and from deterministic (non-trained) agents defined within params.

    Args:
         args (argparse.Namespace): namespace object of parsed arguments via argparse
         params (Dict[str, any]): dictionary defined in config python file
         dirs (OrderedDict[str, str]): dictionary of file/path locations
    """
    # instantiate objects
    instance_generator = params['instance']['type'](**params['instance'])
    eval_writer = tf.summary.FileWriter(os.path.join(dirs['log_dir'], 'eval')) if not args.nlog else None
    if eval_writer is not None:
        tensorboard_text(writer=eval_writer, tag='hyperparams', value=str_iterable(params, 1, markdown_newline=True))

    # scan for backed up hyperparams in root_dir (these are the trained agents to be evaluated)
    filepaths = [os.path.abspath(os.path.realpath(str(f))) for f in Path(params['root_dir']).glob('**/hyperparams*.py')]
    trained_params = []
    # iterate over params, and load each param instance
    for filepath in filepaths:
        _all_files = os.listdir(os.path.dirname(filepath))
        for _file in _all_files:
            if any(_file.endswith(ext) for ext in [".meta", ".pkl", ".pickle"]):
                trained_params.append(load_hyperparameters(config_file_path=filepath, run_mode='train'))
                break
    logging.info("loading following pre-trained models: {}".format(trained_params))

    # instantiate objects
    environments, tf_graphs, agents = OrderedDict(), OrderedDict(), OrderedDict()

    # load pre-trained agents
    for param in trained_params:
        _env = param['environment']['type'](**param['environment'])
        _graph = tf.Graph()
        with _graph.as_default():
            _agent = param['agent']['type'](**param['agent'], state_dim=_env.state_dim, action_dim=_env.action_dim)
            _agent.load()
            _agent.epsilon = 1.0
        _key = id(_agent)
        environments[_key], agents[_key], tf_graphs[_key] = _env, _agent, _graph

    # load default deterministic agents
    for agent_dict in params['agent']:
        _env = params['environment']['type'](**params['environment'])
        _agent = agent_dict['type'](**agent_dict, state_dim=_env.state_dim, action_dim=_env.action_dim)
        _key = id(_agent)
        environments[_key], agents[_key] = _env, _agent

    # prepare outputs
    if not args.nsave:
        _backup_file_path = os.path.join(dirs['output_dir'], dirs['config_file_name']+'.py')        # backup target
        copyfile(src=dirs['config_file_path'], dst=_backup_file_path)                               # copy and log
        logging.info("backed up '{}' as '{}'".format(dirs['config_file_path'], _backup_file_path))

    # let's evaluate
    evaluate(
        instance_generator=instance_generator,
        environments=environments,
        agents=agents,
        eval_writer=eval_writer
    )


def evaluate(instance_generator, environments, agents, eval_writer):
    """Agents are evaluated on their environments by playing instances.

    instance_generator yields an instance for each environment that its respective agent plays on. Each instance
    constitutes as one episode; each environment is being reset for each instance.

    Args:
        instance_generator (pyloa.instance.InstanceGenerator): generator, which iterates instances to evaluate on
        environments (OrdererdDict[int, pyloa.environment.Environment]): envs, which consume instances and return states
            & rewards
        agents (OrdererdDict[int, pyloa.agent.Agent]): agents that play on their respective environments via actions
        eval_writer (tf.summary.FileWriter): tensorboard logging for evaluation
    """
    logging.info("begin evaluating {} on {} instances from {}".format(str(list(agents.values())),
                                                                      len(instance_generator), str(instance_generator)))

    # iterate over instances
    t = tqdm(instance_generator, desc=str(instance_generator), unit='instance')
    for episode, instance in enumerate(t):
        for _, environment in environments.items():
            environment.reset()
            environment.instance = instance

        for key, agent in agents.items():
            agent.play_instance(environment=environments[key], episode=episode, is_train=False, writer=eval_writer)
        logging.info("finished evaluating episode={}".format(episode))



from abc import abstractmethod
import numpy as np
from collections import OrderedDict
from copy import deepcopy


class Graph(object):
    def __init__(self, vertex_colors):
        if type(self) == Graph:
            raise NotImplementedError("{} must be subclassed.".format(self.__class__.__name__))
        self._vertex_colors = vertex_colors.astype(dtype=np.uint8)

    @abstractmethod
    def __getitem__(self, index):
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def __str__(self):
        raise NotImplementedError("Must be implemented in subclass.")

    def __len__(self):
        return len(self._vertex_colors)

    def __iter__(self):
        return self.__next__()

    def __next__(self):
        for vertex in range(len(self)):
            yield vertex

    def _get_predecessors(self, vertex):
        return self[:, vertex]

    def _get_successors(self, vertex):
        return self[vertex]

    def _get_vertex_colors(self):
        return self._vertex_colors

    def _get_vertices(self):
        return set(range(len(self)))

    @abstractmethod
    def add_edge(self, edge):
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def has_edge(self, edge):
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def remove_edge(self, edge):
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def copy(self):
        raise NotImplementedError("Must be implemented in subclass.")

    def get_color(self, vertex):
        return self._vertex_colors[vertex]

    def get_neighbours(self, vertex):
        return self._get_successors(vertex)

    def get_neighbours_colors(self, vertex):
        _neighbour_colors = self._vertex_colors[self[vertex]]
        return _neighbour_colors[_neighbour_colors > 0]

    def get_degree(self, vertex):
        return len(self[vertex])

    def get_max_degree(self):
        return max([self.get_degree(vertex) for vertex in self])

    def is_locally_colorized(self, vertex):
        return self.get_color(vertex) not in self.get_neighbours_colors(vertex)

    def is_colorized(self):
        if 0 in self.coloring:
            return False
        for vertex in self:
            if not self.is_locally_colorized(vertex):
                return False
        return True

    def astype(self, graph_type, dtype=np.uint8):
        if graph_type == NumpyGraph:
            adjacency_matrix = np.zeros(shape=(len(self), len(self)), dtype=dtype)
            for vertex in self:
                adjacency_matrix[vertex][self[vertex]] = 1
            return NumpyGraph(
                adjacency_matrix=adjacency_matrix,
                vertex_colors=self._vertex_colors,
                dtype=dtype
            )
        if graph_type == AdjacencyListGraph:
            dictionary = OrderedDict()
            for vertex in self:
                dictionary[vertex] = list(self[vertex])
            return AdjacencyListGraph(
                dictionary=dictionary,
                vertex_colors=self._vertex_colors
            )
        if graph_type == EdgeListGraph:
            edge_list = []
            for vertex in self:
                for n_vertex in self[vertex]:
                    edge = {vertex, n_vertex}
                    if edge not in edge_list:
                        edge_list.append(edge)
            return EdgeListGraph(
                edge_list=edge_list,
                vertex_colors=self._vertex_colors
            )

    def __eq__(self, other):
        # correct class, correct number of vertices
        if not isinstance(other, Graph) or len(self) != len(other):
            return False
        for vertex in self:
            # vertex equality
            if vertex not in other or self.get_color(vertex) != other.get_color(vertex):
                return False
            # neighbour equality
            for n_vertex in self[vertex]:
                if not other.has_edge({vertex, n_vertex}):
                    return False
        return True

    number_of_vertices = property(fset=None, fget=__len__, fdel=None)
    coloring = property(fset=None, fget=_get_vertex_colors, fdel=None)
    vertices = property(fset=None, fget=_get_vertices, fdel=None)


class AdjacencyListGraph(Graph):
    def __init__(self, dictionary, vertex_colors):
        super(AdjacencyListGraph, self).__init__(vertex_colors=vertex_colors)
        self._dictionary = deepcopy(dictionary)

    def __getitem__(self, index):
        return self._dictionary[index]

    def __repr__(self):
        return repr(self._dictionary)

    def __str__(self):
        return str(self._dictionary)

    def add_edge(self, edge):
        v1, v2 = edge
        self[v1].append(v2)
        self[v2].append(v1)

    def has_edge(self, edge):
        v1, v2 = edge
        return v1 in self[v2] and v2 in self[v1]

    def remove_edge(self, edge):
        v1, v2 = edge
        self[v1].remove(v2)
        self[v2].remove(v1)

    def copy(self):
        return AdjacencyListGraph(dictionary=self._dictionary, vertex_colors=self._vertex_colors)


class EdgeListGraph(Graph):
    def __init__(self, edge_list, vertex_colors):
        super(EdgeListGraph, self).__init__(vertex_colors=vertex_colors)
        self._edge_list = deepcopy(edge_list)

    def __getitem__(self, index):
        return [edge.difference({index}).pop() for edge in self._edge_list if index in edge]

    def __repr__(self):
        return repr(self._edge_list)

    def __str__(self):
        return str(self._edge_list)

    def add_edge(self, edge):
        self._edge_list.append(set(edge))

    def has_edge(self, edge):
        return set(edge) in self._edge_list

    def remove_edge(self, edge):
        self._edge_list.remove(set(edge))

    def copy(self):
        return EdgeListGraph(edge_list=self._edge_list, vertex_colors=self._vertex_colors)


class NumpyGraph(Graph):
    def __init__(self, adjacency_matrix, vertex_colors, dtype=np.uint8):
        super(NumpyGraph, self).__init__(vertex_colors=vertex_colors)
        self._adjacency_matrix = adjacency_matrix.astype(dtype=dtype)
        self._dtype = dtype

    def __getitem__(self, index):
        return np.flatnonzero(self._adjacency_matrix[index])

    def __repr__(self):
        return repr(self._adjacency_matrix)

    def __str__(self):
        return str(self._adjacency_matrix)

    def add_edge(self, edge):
        v1, v2 = edge
        self._adjacency_matrix[([v1, v2], [v2, v1])] = 1

    def has_edge(self, edge):
        v1, v2 = edge
        return self._adjacency_matrix[v1, v2] == self._adjacency_matrix[v2, v1] == 1

    def remove_edge(self, edge):
        v1, v2 = edge
        self._adjacency_matrix[([v1, v2], [v2, v1])] = 0

    def copy(self):
        return NumpyGraph(adjacency_matrix=self._adjacency_matrix, vertex_colors=self._vertex_colors, dtype=self._dtype)



import os
import logging
from shutil import copyfile

import tensorflow as tf
from tqdm import tqdm

from pyloa.utils import str_iterable, tensorboard_text


def run(args, params, dirs):
    """Sets up all necessary objects and containers before training.

    Firstly, instantiates the necessary objects defined in params (agent, instance_generator, environment); secondly.
    backs up the configuration file and starts to train agent.

    Args:
         args (argparse.Namespace): namespace object of parsed arguments via argparse
         params (Dict[str, any]): dictionary defined in config python file
         dirs (OrderedDict[str, str]): dictionary of file/path locations
    """
    # instantiate objects
    instance_generator = params['instance']['type'](**params['instance'])
    environment = params['environment']['type'](**params['environment'])
    eval_writer = tf.summary.FileWriter(os.path.join(dirs['log_dir'], 'eval')) if not args.nlog else None
    train_writer = tf.summary.FileWriter(os.path.join(dirs['log_dir'], 'train')) if not args.nlog else None
    if eval_writer is not None:
        tensorboard_text(writer=eval_writer, tag='hyperparams', value=str_iterable(params, 1, markdown_newline=True))
    agent = params['agent']['type'](**params['agent'], state_dim=environment.state_dim,
                                    action_dim=environment.action_dim)
    _agent_file = os.path.join(dirs['output_dir'], str(agent))

    # prepare outputs
    if not args.nsave:
        _backup_file_path = os.path.join(dirs['output_dir'], dirs['config_file_name']+'.py')        # backup target
        copyfile(src=dirs['config_file_path'], dst=_backup_file_path)                               # copy and log
        logging.info("backed up '{}' as '{}'".format(dirs['config_file_path'], _backup_file_path))

    # let's train
    train(
        instance_generator=instance_generator,
        environment=environment,
        agent=agent,
        eval_writer=eval_writer,
        train_writer=train_writer,
        checkpoint_step=params['checkpoint_step'] if not args.nsave else 0,
        agent_file=_agent_file
    )


def train(instance_generator, environment, agent, eval_writer, train_writer, checkpoint_step, agent_file):
    """Agent learns his environment by playing instances.

    instance_generator yields an instance for the environment that the agent plays on. Each instance constitutes as one
    episode; the environment is being reset for each instance. The agent is evaluated and afterwards trained on each
    instance. A checkpoint is saved every checkpoint_step-many episodes.

    Args:
        instance_generator (pyloa.instance.InstanceGenerator): iterates instances to train on
        environment (pyloa.environment.Environment): consumes instances and returns states & rewards
        agent (pyloa.agent.Agent): plays/trains on environment via actions
        eval_writer (tf.summary.FileWriter): tensorboard logging for evaluation
        train_writer (tf.summary.FileWriter): tensorboard logging for training
        checkpoint_step (int): defines after how many episodes a checkpoint of agent is saved
        agent_file (str): file location of agent to be stored at
    """
    logging.info("begin training {} on {} instances from {}".format(agent, len(instance_generator), instance_generator))
    with tqdm(instance_generator, desc=str(agent), unit='instance') as t:
        for episode, instance in enumerate(t):
            # start a new episode
            environment.reset()
            environment.instance = instance

            # eval on new instance
            agent.play_instance(environment=environment, is_train=False, episode=episode, writer=eval_writer)

            # reset, train on same instance
            environment.reset()
            environment.instance = instance

            # train on new instance
            agent.play_instance(environment=environment, is_train=True, episode=episode, writer=train_writer)

            # save step
            if checkpoint_step and (episode + 1) % checkpoint_step == 0:
                agent.save(agent_file)

    logging.info("finished training {}".format(agent))



class ConfigurationException(RuntimeError):
    def __init__(self, msg=None):
        if type(self) == ConfigurationException:
            raise NotImplementedError("{} must be subclassed.".format(self.__class__.__name__))
        super().__init__(msg)


class HyperparamsConfigurationException(ConfigurationException):
    def __init__(self, msg=None):
        super().__init__(msg)


class DirectoryNotFoundError(OSError):
    def __init__(self, msg=None):
        super().__init__(msg)


class InvalidFileType(OSError):
    def __init__(self, msg=None):
        super().__init__(msg)


class InvalidDirectoryLocation(OSError):
    def __init__(self, msg=None):
        super().__init__(msg)



import numpy as np

from .instancegenerator import InstanceGenerator


class SequenceGenerator(InstanceGenerator):
    """A generator to iteratively generate sequence_number-many sequences of size sequence_size.

    Args:
        sequence_size (int): target size of each sequence
        sequence_number (int): number of sequences to generate till generator depletes
        min_page (int): minimum element within each generated sequence
        max_page (int): maximum element within each generated sequence
        dtype (type): type of each element within sequence

    Attributes:
        _sequence_size (int): target size of each sequence
        _min_page (int): minimum element within each generated sequence
        _max_page (int): maximum element within each generated sequence
        _num_pages (int): number of different pages that can occur in each generated sequence
        _dtype (type): type of each element within sequence
        _pages (np.ndarray, np.shape(_num_pages,), dtype=_dtype): lookup list for each page that can occur

    Raises:
        AssertionError: if sequence_size, sequence_number or min_page does not exceed 0 or if max_page exceeds min_page
        NotImplementedError: if abstract class is to be instanced
    """
    def __init__(self,
                 sequence_size,
                 sequence_number,
                 min_page,
                 max_page,
                 dtype=np.int32):
        assert sequence_size > 0, "sequence_size must exceed 0"
        assert sequence_number > 0, "sequence_number must exceed 0"
        assert min_page > 0, "min_page must exceed 0"
        assert max_page > min_page, "max_page must exceed min_page"
        super(SequenceGenerator, self).__init__(instance_number=sequence_number)
        self._sequence_size = sequence_size
        self._min_page = min_page
        self._max_page = max_page
        self._num_pages = max_page - min_page + 1
        self._dtype = dtype
        self._pages = np.arange(self._num_pages, dtype=self._dtype) + self._min_page

    def _get_pages(self):
        """np.ndarray: ascending array of pages self can populate sequences with"""
        return self._pages

    def _get_sequence_size(self):
        """int: number of pages per sequence"""
        return self._sequence_size

    pages = property(fset=None, fget=_get_pages, fdel=None)
    sequence_size = property(fset=None, fget=_get_sequence_size, fdel=None)



# define abstract alias imports for other modules
from .instancegenerator import InstanceGenerator
from .seqgen import SequenceGenerator
from .graphgenerator import GraphGenerator
# define alias imports for other modules
from .randomseqgen import RandomSequenceGenerator, RandomDistributionSequenceGenerator
from .markovseqgen import MarkovChainSequenceGenerator
from .graphgenerator import RandomGraphGenerator, DeltaRandomGraphGenerator



from collections.abc import Iterable
import logging

import numpy as np


class SumTree(object):
    """A Fenwick (Sum)Tree implementation used as an indexing structure. This implementation assures numerical stability
    via auto-correction. This implementation is not threadsafe.

    SumTree can hold _memory_size many data entries, which are contained in _data as ring-buffer implementation. _tree
    is an array based implementation of a tree and depicts the SumTree used to index _data. Each leaf of _tree  indexes
    a data-entry instance. The value of each leaf in _tree depicts a data-entries' priority. Each non-leaf node in _tree
    is the sum of its children.

        Example:
            Assume a memory_size = 8 instance:

                _tree:
                                            24                          # sum of children's priorities
                                          /   \
                                         /     \
                                        /       \
                                       /         \
                                     13          11                     # sum of children's priorities
                                     / \         / \
                                    /   \       /   \
                                   4     9     5     6                  # sum of children's priorities
                                  / \   / \   / \   / \
                                 1  3  2  7  3  2  4  2                 # leafs; priority of each data entry below
                _data:
                                [a, b, c, d, e, f, g, h]                # coresponding data entries

                _tree as array:
                    [24, 13, 11, 4, 9, 5, 6, 1, 3, 2, 7, 3, 2, 4, 2]    # _tree as actual array


            To index _memory_size many data entries _tree needs to be of size: _memory_size * 2 - 1. A data_index in
            _data corresponds to the following tree_index:

                    tree_index = data_index + _memory_size - 1

                    <=>

                    data_index = tree_index - _memory_size + 1

            In this case the data_index of 'c' is 2; its tree_index is 2 + 8 - 1 = 9 and therefore its current priority
            is _tree[9] = 2. Updating the priority of an existing data entry is propagated upwards throughout _tree in
            O(log(2 * _memory_size - 1)). Assume invocation of update(tree_index=9, priority=1):

                _tree:
                                            23                      # sum of children's priorities
                                          /   \
                                         /     \
                                        /       \
                                       /         \
                                      12         11                 # sum of children's priorities
                                     / \         / \
                                    /   \       /   \
                                   4     8     5     6              # sum of children's priorities
                                  / \   / \   / \   / \
                                 1  3  1  7  3  2  4  2             # leafs; priority of each data entry below

            The update is, depending on the size of the tree and the range of _tree values, not always numerically
            stable. This means that _tree[0] = sum(_tree[-_memory_size:]) will eventually not hold to be true, if
            floating point precision is exceeded when relatively small priority updates are performed on a tree_index.

    Therefore, update is not always numerically stable; Priority updates within the index structure can result in small
    delta updates, which can not always be propagated upwards throughout the entire tree due to floating point precision
    limitations; the SumTree implementation at hand checks for "failing" delta updates and auto-corrects "too small
    delta changes" to the next applicable machine epsilon. If the auto-correct exceeds a relative change of
    _max_relative_change with said auto-correction, a warning is logged. SumTree is agnostic with regards to its data
    entries.

    Args:
        memory_size (int): number of data entries to store and index
        max_relative_change (number): maximum relative change of SumTree's auto-correction for numerical stability
        dtype (type): floating point precision used in SumTree

    Attributes:
        _memory_size (int): number of data entries to store and index
        _data_cursor (int): current cursor position
        _data (np.ndarray): container for data entries
        _tree_size (int): number of nodes needed to index _memory_size many data entries
        _max_relative_change (number): maximum relative change of SumTree's auto-correction for numerical stability
        _dtype (type): dtype for floating point precision used in SumTree
        _tree (np.ndarray): container for tree nodes
        _logger (logging.Logger): general purpose logger
    """
    def __init__(self, memory_size, max_relative_change, dtype):
        self._memory_size = memory_size
        self._data_cursor = 0
        self._data = np.zeros(shape=self._memory_size, dtype=object)
        self._tree_size = 2 * self._memory_size - 1
        self._max_relative_change = max_relative_change
        self._dtype = dtype
        self._tree = np.zeros(shape=(self._tree_size,), dtype=self._dtype)
        self._logger = logging.getLogger(__name__)
        self._logger.info("initializing {}".format(self.__class__.__name__))

    def __getitem__(self, index):
        """tuple(type(data)): data entries indexed in _tree with index"""
        if isinstance(index, Iterable):
            return tuple(map(self._get_item, index))
        return self._get_item(index),

    def __len__(self):
        """int: maximum number of data entries SumTree can hold"""
        return self._memory_size

    def add(self, priority, data_entry):
        """Places data_entry at cursor location in _data. Propagates new priority through tree. Increments cursor.
        Operation performs in  O(log(_memory_size)).
        
        Args:
            priority (number): initial priority value of data_entry
            data_entry (object): data_entry to index
        """
        _tree_index = self._data_cursor + len(self) - 1
        self._data[self._data_cursor] = data_entry
        self.update(_tree_index, priority)
        self._data_cursor = (self._data_cursor + 1) % len(self)

    def update(self, tree_index, priority):
        """Propagate a priority update for tree_index throughout tree. Operation performs in  O(log(_memory_size)).

        Calculates the _delta change to propagate through tree. If _delta can be applied to _tree[0], than this
        operation is numerically stable and no auto-correction is needed. If _delta can't be applied to _tree[0], due to
        _delta < np.spacing(_tree[0]), then _delta is auto-corrected to np.spacing(_tree[0]) (smallest possible machine
        epsilon to propagate change throughout SumTree). If _max_relative_change is exceeded, a warning is logged.

        Args:
            tree_index (int): index of leaf to update priority for
            priority (number): new priority

        Raises:
            AssertionError: tree_index is too small and therefore indexes a non-leaf node
        """
        # get max priority from tree, calculate delta change to propagate through SumTree
        assert tree_index >= len(self) - 1, "tree_index='{}' is not indexing a leaf".format(tree_index)
        _max_priority = self._tree[0]
        _delta = priority - self._tree[tree_index]

        # assure numerical stability, log warning if auto-correct is too excessive
        if _delta != 0.0 and _max_priority + _delta == _max_priority:       # floating precision exceeded
            _delta = np.sign(_delta) * np.spacing(_max_priority)            # minimum eps to apply change to tree
            _relative_change = _delta/(priority - self._tree[tree_index])   # relative change to _delta
            self._logger.debug("SumTree delta update changed from '{}' to '{}' (relative_change='{:5.5f}') due to "
                               "numerical stability".format(priority-self._tree[tree_index], _delta, _relative_change))
            if _relative_change > self._max_relative_change:                # relative change too excessive
                _warn_msg = "SumTree delta update propagation exceeds floating point precision for dtype='{}': " \
                            "normalize/clip td_errors and rewards OR allow max_relative_change='{}' to be larger"\
                    .format(self._dtype, self._max_relative_change)
                self._logger.warning(_warn_msg)

        # set current priority, propagate change through tree
        self._tree[tree_index] = priority
        while tree_index != 0 and _delta != 0.0:
            tree_index = (tree_index - 1) // 2
            self._tree[tree_index] += _delta

    def clear(self):
        """Clear containers and structures."""
        del self._data[:]
        self._data = np.zeros(shape=self._memory_size, dtype=object)
        self._data_cursor = 0
        self._tree.fill(0.0)

    def _get_priorities(self):
        """np.ndarray: priorities of data entries"""
        return self._tree[-self._memory_size:]

    def _get_total_priority(self):
        """number: prirority of root node (summed priority of all data entries)"""
        return self._tree[0]

    def _get_max_priority(self):
        """number: highest priority of a leaf node"""
        return np.max(self.priorities)

    def _get_item(self, item):
        """Helper function to retrieve an entry from SumTree with index item.

        Args:
            item (number): selector to index a single data entry with

        Returns:
            tuple of (int, number, type(data)): First item is the tree_index, at which selected entry is indexed with.
                Second item is the priority of corresponding the data entry. Third item is the corresponding data entry.
        """
        _parent = 0

        while _parent < self._memory_size - 1:
            _left_child = 2 * _parent + 1
            _right_child = _left_child + 1
            if item <= self._tree[_left_child]:
                _parent = _left_child
            else:
                item -= self._tree[_left_child]
                _parent = _right_child

        _data_index = _parent - len(self) + 1
        return _parent, self._tree[_parent], self._data[_data_index]

    def _get_data_size(self):
        """int: number of data entries stored"""
        return len(self._data[self._data != 0])

    data_size = property(fset=None, fget=_get_data_size, fdel=None)
    priorities = property(fset=None, fget=_get_priorities, fdel=None)
    max_priority = property(fset=None, fget=_get_max_priority, fdel=None)
    total_priority = property(fset=None, fget=_get_total_priority, fdel=None)



import os
import sys
import logging
from importlib.util import find_spec, module_from_spec

from .format import str_iterable
from .exception import HyperparamsConfigurationException
from pyloa.instance import InstanceGenerator, SequenceGenerator
from pyloa.agent import Agent, RLAgent, PagingAgent
from pyloa.environment import Environment, PagingEnvironment


def load_hyperparameters(config_file_path, run_mode):
    """A helper function that imports a hyperparameter python-file and checks its dictionary 'params' for validity.

    Args:
        config_file_path (str): path to python-file
        run_mode (str): run mode of project

    Returns:
        dict: syntactically correct hyperparameters for run mode

    Raises:
        ModuleNotFoundError: if file can't be imported
        HyperparamsConfigurationException: if syntax of config-file is invalid (missing values, incorrect types, etc.)
    """
    config_dir, config_file = os.path.split(config_file_path)
    config_file_name, config_file_type = config_file.rsplit(".", 1)

    # update sys path, load from spec
    sys.path.append(config_dir)
    spec = find_spec(config_file_name)
    if spec is None:
        raise ModuleNotFoundError("Can not import module '{}' via spec")
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    sys.path.remove(config_dir)
    params = module.params

    # check params for completeness and validity
    if not isinstance(params, dict):
        raise HyperparamsConfigurationException("params in config_file='{}' must be of type='{}'".format(
            config_file_path, dict))
    if 'instance' not in params:
        raise HyperparamsConfigurationException("instance must be specified in params")
    if 'type' not in params['instance']:
        raise HyperparamsConfigurationException("type of instance must be specified in params")
    if not issubclass(params['instance']['type'], InstanceGenerator):
         raise HyperparamsConfigurationException("instance must be subclassed of '{}'".format(InstanceGenerator))
    if 'environment' not in params:
        raise HyperparamsConfigurationException("environment must be specified in params")
    if 'type' not in params['environment']:
        raise HyperparamsConfigurationException("type of environment must be specified in params")
    if not issubclass(params['environment']['type'], Environment):
        raise HyperparamsConfigurationException("environment must be subclassed of '{}'".format(Environment))
    if 'agent' not in params:
        raise HyperparamsConfigurationException("agent must be specified in params")

    # run mode specific entries
    if run_mode == 'train':
        if 'type' not in params['agent']:
            raise HyperparamsConfigurationException("type of agent must be specified in params")
        if not issubclass(params['agent']['type'], RLAgent):
            raise HyperparamsConfigurationException("agent must be subclassed of '{}'".format(RLAgent))
        if 'checkpoint_step' not in params:
            raise HyperparamsConfigurationException("checkpoint_step must be specified in params")
        if not isinstance(params['checkpoint_step'], int) or not params['checkpoint_step'] >= 0:
            raise HyperparamsConfigurationException("checkpoint_step must be unsigned int")
    if run_mode == 'eval':
        if not isinstance(params['agent'], list):
            raise HyperparamsConfigurationException("agent must be a list of agents to evaluate")
        if len(params['agent']) == 0:
            raise HyperparamsConfigurationException("list of agents to evaluate must contain at least one item")
        for agent in params['agent']:
            if 'type' not in agent:
                raise HyperparamsConfigurationException("type of agent='{}' must be specified in params".format(agent))
            if not issubclass(agent['type'], Agent):
                raise HyperparamsConfigurationException("agent must be subclassed of '{}'".format(Agent))
        if 'root_dir' not in params:
            raise HyperparamsConfigurationException("root_dir must be specified to evaluate learned agents")
    if run_mode == 'gen':
        if 'type' not in params['agent']:
            raise HyperparamsConfigurationException("type of agent must be specified in params")
        if not issubclass(params['agent']['type'], PagingAgent):
            raise HyperparamsConfigurationException("agent must be subclassed of '{}'".format(PagingAgent))
        if not issubclass(params['environment']['type'], PagingEnvironment):
            raise HyperparamsConfigurationException("environment must be subclassed of '{}'".format(PagingEnvironment))
        if not issubclass(params['instance']['type'], SequenceGenerator):
            raise HyperparamsConfigurationException("instance must be subclassed of '{}'".format(SequenceGenerator))

    # log and ship
    logging.info("params from {} successfully loaded".format(config_file_path))
    logging.debug("params={}".format(str_iterable(params)))
    return params



import os
import logging
from collections import OrderedDict
from shutil import copyfile

import tensorflow as tf
from tqdm import tqdm
import numpy as np

from pyloa.agent import BeladysOptimalAgent, RLAgent
from pyloa.utils import tensorboard_array

count_index = 0


def run(args, params, dirs):
    """Sets up all necessary objects and containers before genetic algorithm is envoked to evaluate agent.

    Firstly, instantiates the necessary objects defined in params (agent, sequence_generator, environment); secondly.
    backs up the configuration file and starts to eval agent. Agent, Environment and Generator have to be a "paging
    problem instance" for genetic evaluation (vertex problem is NOT applicable for this runmode).

    Args:
        args (argparse.Namespace): namespace object of parsed arguments via argparse
        params (Dict[str, any]): dictionary defined in config python file
        dirs (OrderedDict[str, str]): dictionary of file/path locations
    """
    # instantiate objects
    sequence_generator = params['instance']['type'](**params['instance'])
    environment = params['environment']['type'](**params['environment'])
    eval_writer = tf.summary.FileWriter(os.path.join(dirs['log_dir'], 'eval')) if not args.nlog else None
    agent = params['agent']['type'](**params['agent'], state_dim=environment.state_dim,
                                    action_dim=environment.action_dim)

    # load agent from save_file, set epsilon to 1.0 (pure evaluation)
    if isinstance(agent, RLAgent):
        agent.load()
        agent.epsilon = 1.0

    # agent container
    agents = OrderedDict()
    belady_agent = BeladysOptimalAgent(state_dim=environment.state_dim, action_dim=environment.action_dim)
    agents[id(belady_agent)] = belady_agent
    agents[id(agent)] = agent

    # prepare outputs
    if not args.nsave:
        _backup_file_path = os.path.join(dirs['output_dir'], dirs['config_file_name']+'.py')        # backup target
        copyfile(src=dirs['config_file_path'], dst=_backup_file_path)                               # copy and log
        logging.info("backed up '{}' as '{}'".format(dirs['config_file_path'], _backup_file_path))

    # evaluate
    diversification_size = params['diversification_size'] if 'diversification_size' in params else \
        len(sequence_generator.pages)**2
    evaluate(
        sequence_size=sequence_generator.sequence_size,
        pages=sequence_generator.pages,
        cache_size=environment.cache_size,
        window_size=params['window_size'] if 'window_size' in params else environment.cache_size,
        agents=agents,
        environment=environment,
        eval_writer=eval_writer,
        diversification_size=diversification_size,
        start_size=params['start_size'] if 'start_size' in params else 20,
        selection_size=params['selection_size'] if 'selection_size' in params else 5,
    )


def generate_start_individuum(pages, sequence_size, window_size):
    # Todo @schoenitz documentation
    seq = np.full(sequence_size, np.random.choice(pages, 1))
    seq[0:window_size] = create_window_mutation(pages, window_size)
    return seq


def create_window_mutation(pages, w_size, replace=False, ignore_as_first_page=None):
    # Todo @schoenitz documentation
    if not replace and w_size > len(pages):
        raise ValueError("window size must not exceed the number of pages")
    if ignore_as_first_page is None:
        return np.random.choice(pages, w_size, replace=replace)
    else:
        return np.append(
            np.random.choice([page for page in pages if page != ignore_as_first_page], 1),
            np.random.choice(pages, w_size - 1, replace=replace)
        )


def index_branching(sequence, pages, window_size, index):
    # Todo @schoenitz documentation
    branched_sequences = []
    for relative_idx in range(window_size):
        for page in pages:
            branched_sequences.append(np.copy(sequence))
            branched_sequences[-1][index + relative_idx] = page
    return branched_sequences


def genetic_mutation(population, pages, diversification_size, window_size, index):
    # Todo @schoenitz documentation
    _new_population = []
    for individuum in population:
        for _ in range(diversification_size):
            _tmp = np.copy(individuum)
            _new_window = create_window_mutation(pages, window_size, ignore_as_first_page=_tmp[index - 1])
            _tmp[index:index + window_size] = _new_window
            _new_population.append(_tmp)
    return _new_population


def genetic_branching(population, pages, window_size, index):
    # Todo @schoenitz documentation
    _new_population = []
    for individuum in population:
        _new_population.extend(index_branching(sequence=individuum, pages=pages, window_size=window_size, index=index))
    return _new_population


def genetic_selection(sequences, scores, selection_size):
    # Todo @schoenitz documentation
    _args_top_scores = np.argsort(scores)[-selection_size:][::-1]
    _best_individuums = [sequences[arg] for arg in _args_top_scores]
    _best_scores = [scores[arg] for arg in _args_top_scores]
    return _best_individuums, _best_scores


def evaluate_population(agents, environment, population, eval_writer):
    # Todo @schoenitz documentation
    global count_index
    _approximations = []
    _len = len(population)

    for idx, sequence in enumerate(population):
        _agent_steps, _belady_steps = 0, 0
        _agent_faults, _belady_faults = 0, 0
        _agent_costs, _belady_costs = 0, 0
        # reset envs, set instance to each env
        for key, agent in agents.items():
            environment.reset()
            environment.instance = sequence

            # evaluate agent on env
            _step, _cost, _fault, = agent.play_instance(environment=environment, writer=eval_writer,
                                                        episode=count_index, is_train=False)
            if isinstance(agent, BeladysOptimalAgent):
                _belady_steps += _step
                _belady_costs += _cost
                _belady_faults += _fault
            else:
                _agent_steps += _step
                _agent_costs += _cost
                _agent_faults += _fault
                _approx = 1.0 if _belady_faults == 0 else _agent_faults / _belady_faults
                _approximations.append(_approx)
                tensorboard_array(eval_writer, 'approx/'+str(agent), _approximations, count_index)
        count_index += 1
    return _approximations


def evaluate(sequence_size, pages, cache_size, window_size, agents, environment, eval_writer, diversification_size,
             start_size=25, selection_size=5):
    """Performs a genetic algorithm on agents to calculate the competitive ratio of agent vs. BeladysOptimalAgent.

    #Todo @schoenitz LONG DESCRIPTION

    Args:
        sequence_size (int): size of sequences
        pages (np.ndarray): pages that each sequence may contain
        cache_size (int): cache_size of paging problem
        window_size (int): window_size that genetic may use to alter a sequence
        agents (OrderedDict[int, pyloa.agent.PagingAgent]): two agent instances (Belady and testee) to evaluate
        environment (pyloa.environment.PagingEnvironment): environment to evluate on
        eval_writer (tf.summary.FileWriter): tensorboard logging for evaluation
        diversification_size (int): number of diversifications per individuum during mutation step
        start_size (int): size of initial start population
        selection_size (int): number of "best performing individuums" to keep during selection steo
    """
    # vars
    episodes = int((sequence_size-cache_size)/window_size)
    gen = (cache_size + window_size * i for i in range(episodes))
    current_population = [generate_start_individuum(pages, sequence_size, window_size) for _ in range(start_size)]

    logging.info("starting genetic algorithm evaluation with window_size={}, diversification_size={}, start_size={}, "
                 "selection_size={}".format(window_size, diversification_size, start_size, selection_size))

    t = tqdm(iterable=enumerate(gen), desc='GeneticAlgorithm', total=episodes+1)
    for episode, window_index in t:
        # eval step
        scores = evaluate_population(agents, environment, current_population, eval_writer)
        logging.info("metrics after branch-{}: population_size={}, mean={:3.3f}, max={:3.3f}, min={:3.3f}, std={:3.3f}"
                     .format(episode, len(scores), np.mean(scores), np.max(scores), np.min(scores), np.std(scores)))

        # select best performing individuums, wipe the rest
        current_population, scores = genetic_selection(current_population, scores, selection_size)
        logging.info("metrics after select-{}: population_size={}, mean={:3.3f}, max={:3.3f}, min={:3.3f}, std={:3.3f}"
                     .format(episode, len(scores), np.mean(scores), np.max(scores), np.min(scores), np.std(scores)))

        # mutation step
        current_population = genetic_mutation(current_population, pages, diversification_size, window_size, window_index)

        # eval step
        scores = evaluate_population(agents, environment, current_population, eval_writer)
        logging.info("metrics after mutate-{}: population_size={}, mean={:3.3f}, max={:3.3f}, min={:3.3f}, std={:3.3f}"
                     .format(episode, len(scores), np.mean(scores), np.max(scores), np.min(scores), np.std(scores)))

        # select best performing individuums, wipe the rest
        current_population, scores = genetic_selection(current_population, scores, selection_size)
        logging.info("metrics after select-{}: population_size={}, mean={:3.3f}, max={:3.3f}, min={:3.3f}, std={:3.3f}"
                     .format(episode, len(scores), np.mean(scores), np.max(scores), np.min(scores), np.std(scores)))

        # branching step
        current_population = genetic_branching(current_population, pages, window_size, window_index)

    # final eval step
    scores = evaluate_population(agents, environment, current_population, eval_writer)
    final_population, scores = genetic_selection(current_population, scores, selection_size)
    logging.info("final average genetic score: {:3.3f}".format(np.mean(scores)))
    logging.info("Top-{} performing individuums:".format(selection_size))
    for individuum in final_population:
        logging.info("{}".format(individuum))
    logging.info("genetic algorithm finished!")



import numpy as np

from .seqgen import SequenceGenerator


class MarkovChainSequenceGenerator(SequenceGenerator):
    """A SequenceGenerator that generates sequence_number many sequences of length sequence_size.

    A transition_matrix is initially rolled with transition_matrix = uniform(sigma, mu) of shape (num_pages, num_pages);
    whereas the j-th row of transition_matrix depicts the probability_dsitribution of page j. Afterwards softmax is
    applied on each row of transition_matrix. Pages are picked from the interval [min_page, max_page] according to the
    softmaxed probability_distribution of the last page picked. The initial last page to start generating a sequence is
    picked equally distributed.

    Args:
        sequence_size (int): target size of each sequence
        sequence_number (int): number of sequences to generate till generator depletes
        min_page (int): minimum element within each generated sequence
        max_page (int): maximum element within each generated sequence
        sigma (float): sigma for uniform(sigma, mu) probability distribution
        mu (float): mu for uniform(sigma, mu) probability distribution
        dtype (type): type of each element within sequence,
        seed (int): seed for random number generation

    Attributes:
        _rnd_state (np.random.RandomState): random number generator for self with own local seed
        _seed (int or None): seed _rnd_state was initialized with
        _sigma (float): sigma for uniform(sigma, mu) probability distribution
        _mu (float): mu for uniform(sigma, mu) probability distribution
        _transition_matrix (np.npdarray, shape=(_num_pages,_num_pages), dtype=np.float32): a matrix where the j-th row
            depicts j-th page's softmaxed uniform(sigma, mu) distribution to transition with to following pages
    """
    # noinspection PyUnusedLocal
    def __init__(self,
                 sequence_size,
                 sequence_number,
                 min_page,
                 max_page,
                 sigma=1,
                 mu=0,
                 dtype=np.int32,
                 seed=None,
                 **kwargs):
        super(MarkovChainSequenceGenerator, self).__init__(
            sequence_size=sequence_size,
            sequence_number=sequence_number,
            min_page=min_page,
            max_page=max_page,
        )
        self._rnd_state = np.random.RandomState(seed=seed)
        self._seed = seed
        self._sigma = sigma
        self._mu = mu
        self._generate_matrix()
        self._logger.info("initialized {} with transition_matrix='{}'".format(self, self._transition_matrix))

    def _generate_matrix(self):
        """rolls the _transition_matrix with uniform(sigma, mu) and applies a softmax to each row"""
        # roll transition matrix
        self._transition_matrix = self._sigma * self._rnd_state.randn(self._num_pages, self._num_pages) + self._mu
        # apply softmax normalization to each row
        for idx, val in enumerate(self._transition_matrix):
            _exp = np.exp(val)
            self._transition_matrix[idx, :] = _exp / np.sum(_exp)

    def _generate_instance(self):
        """Create and return a newly generated sequence.

        This function initially picks a last_page equally distributed; based on last_page's transition (row entry) the
        following page in sequence is selected, which once again will become last_page. This process is repeated untill
        the sequence is of size sequence_size.

        Returns:
            np.ndarray: sequence with shape=(_sequence_size,) and dtype=_dtype generated from _transition_matrix
        """
        res = np.empty(self._sequence_size, dtype=self._dtype)
        # randomly roll equally distributed what the initial page is
        _last_page = self._rnd_state.randint(low=self._min_page, high=self._max_page+1)
        for idx in range(self._sequence_size):
            # pick next page from _last_page's transitions
            res[idx] = self._rnd_state.choice(self._pages, size=1, p=self._transition_matrix[_last_page-1])
            _last_page = res[idx]
        return res

    def _get_transition_matrix(self):
        """np.ndarray: matrix to draw sequences from with shape=(_num_pages,_num_pages) and dtype=np.float32"""
        return self._transition_matrix

    def _set_transition_matrix(self, value):
        assert isinstance(value, np.ndarray), "transition_matrix must be of type='{}'".format(np.ndarray)
        assert value.shape == (self._num_pages, self._num_pages), "transition_matrix must be of shape='{}' but " \
            "received shape='{}'".format((self._num_pages, self._num_pages), value.shape)
        for idx, row in enumerate(value):
            assert np.sum(row) == 1.0, "summed row='{}' at index='{}' must equal 1.0".format(row, idx)
        self._transition_matrix = value

    def __str__(self):
        """str: string representation of self"""
        return self.__class__.__name__ + "(seed={}, \u03C3={}, \u03BC={})".format(self._seed, self._sigma, self._mu)

    transition_matrix = property(fset=_set_transition_matrix, fget=_get_transition_matrix, fdel=None)




import os
from copy import deepcopy

import numpy as np
import matplotlib.pyplot as plt

from pyloa.utils.graph import Graph, NumpyGraph, EdgeListGraph, AdjacencyListGraph
from pyloa.utils.exception import DirectoryNotFoundError
from .instancegenerator import InstanceGenerator


class GraphGenerator(InstanceGenerator):
    """Abstract base class for to iteratively generate graph_number-many graphs of size graph_size.

    Args:
        graph_type (type): graph type (implementations of pyloa.utils.Graph) that self shall return
        graph_number (int): how many graph instances self will return before depletion
        graph_size (int): maximum size of graph (vertex count) that self will return

    Attributes:
        _graph_type (type): graph type (implementations of pyloa.utils.Graph) that self shall return
        _graph_size (int): maximum size of graph (vertex count) that self will return
        _vertices (set): vertices each graph has to contain

    Raises:
        AssertionError: if graph_number, graph_size does not exceed 0 or if graph_type is not of appropriate type
        NotImplementedError: if abstract class is to be instanced
    """
    def __init__(self,
                 graph_type,
                 graph_number,
                 graph_size):
        assert graph_number > 0, "graph_number must exceed 0 but received {} instead".format(graph_number)
        assert graph_size > 0, "graph_size must exceed 0 but received {} instead".format(graph_size)
        assert issubclass(graph_type, Graph), "graph_type must be subclass of {}".format(Graph)
        super(GraphGenerator, self).__init__(instance_number=graph_number)
        self._graph_type = graph_type
        self._graph_size = graph_size
        self._vertices = {v for v in range(self._graph_size)}

    def _get_vertices(self):
        """set of int: vertices each generated graph instance will contain"""
        return deepcopy(self._vertices)

    def _get_graph_size(self):
        """int: number of vertices each generated graph instance will consist of"""
        return self._graph_size

    def _get_graph_type(self):
        """type: graph_type that self generates"""
        return self._graph_type

    def _set_graph_type(self, value):
        """Set _graph_type.

        Args:
            value (type): graph_type that self generates

        Raises:
            AssertionError: if value is not appropiate type (subclass implementation of Graph)
        """
        assert isinstance(value, Graph), "graph_type must subclass {} but found {} instead".format(Graph, type(value))
        self._graph_type = value

    vertices = property(fset=None, fget=_get_vertices, fdel=None)
    graph_size = property(fset=None, fget=_get_graph_size, fdel=None)
    graph_type = property(fset=_set_graph_type, fget=_get_graph_type, fdel=None)


class RandomGraphGenerator(GraphGenerator):
    """A GraphGenerator that generates graph_number-many random NumpyGraphs of size graph_size. The initial coloring
    of each graph instance is undefined (all vertices have color 0).

    Args:
        graph_number (int): how many graph instances self will return before depletion
        graph_size (int): exact size of graph (vertex count) that self will return
        dtype (type): dtype used for NumpyGraph's adjacency_matrix
        seed (int): seed for random number generation

    Attributes:
        _seed (int or None): seed _rnd_state was initialized with
        _rnd_state (np.random.RandomState): random number generator for self with own local seed
        _dtype (type): dtype used for NumpyGraph's adjacency_matrix
    """
    def __init__(self,
                 graph_number,
                 graph_size,
                 dtype=np.uint8,
                 seed=None,
                 **kwargs):
        super(RandomGraphGenerator, self).__init__(
            graph_type=NumpyGraph,
            graph_number=graph_number,
            graph_size=graph_size,
        )
        self._rnd_state = np.random.RandomState(seed=seed)
        self._seed = seed
        self._dtype = dtype
        self._logger.info("initialized {}".format(self))

    def _generate_instance(self):
        """NumpyGraph: newly generated graph instance"""
        # randomly roll a matrix
        rnd_matrix = self._rnd_state.randint(0, 2, size=(self._graph_size, self._graph_size), dtype=self._dtype)
        # use upper triangle to enforce direction and set diagonal to 0
        rnd_matrix -= np.tril(rnd_matrix, 0)
        rnd_matrix += np.triu(rnd_matrix, 1).transpose()
        return self.graph_type(adjacency_matrix=rnd_matrix, vertex_colors=np.zeros(self._graph_size, dtype=self._dtype))

    def __str__(self):
        """str: string representation of self"""
        return self.__class__.__name__ + "(seed={})".format(self._seed)


class DeltaRandomGraphGenerator(GraphGenerator):
    """A GraphGenerator that generates graph_number-many random NumpyGraphs of size graph_size where for each vertex of
    a generated graph instance degree(vertex) <= max_delta holds to be true. The initial coloring of each graph instance
    is undefined (all vertices have color 0).

    Args:
        max_delta (int): maximum degree that each vertex of generated graphs are allowed to have
        graph_number (int): how many graph instances self will return before depletion
        graph_size (int): exact size of graph (vertex count) that self will return
        dtype (type): dtype used for NumpyGraph's adjacency_matrix
        seed (int): seed for random number generation

    Attributes:
        _max_delta (int): maximum degree that each vertex of generated graphs are allowed to have
        _seed (int or None): seed _rnd_state was initialized with
        _rnd_state (np.random.RandomState): random number generator for self with own local seed
        _dtype (type): dtype used for NumpyGraph's adjacency_matrix
    """
    def __init__(self,
                 max_delta,
                 graph_number,
                 graph_size,
                 dtype=np.uint8,
                 seed=None,
                 **kwargs):
        assert max_delta >= 2, "max_delta must exceed 1 but received {} instead".format(max_delta)
        super().__init__(
            graph_type=NumpyGraph,
            graph_number=graph_number,
            graph_size=graph_size,
        )
        self._max_delta = max_delta
        self._rnd_state = np.random.RandomState(seed=seed)
        self._seed = seed
        self._dtype = dtype
        self._heatmap = np.zeros(shape=(self._graph_size, self._graph_size), dtype=np.float64)
        self._count = 0
        self._logger.info("initialized {}".format(self))

    def _generate_instance(self):
        """NumpyGraph: newly generated graph instance"""
        adjacency_matrix = np.zeros(shape=(self._graph_size, self._graph_size), dtype=self._dtype)
        degrees = self._rnd_state.randint(low=0, high=self._max_delta + 1, size=self.graph_size)

        _amax = np.amax(degrees)
        arg_maxes = np.argwhere(degrees == _amax).flatten()
        idx = arg_maxes[self._rnd_state.randint(len(arg_maxes))]
        max_degree = degrees[idx]
        if max_degree == 0:
            self._update_heatmap(adjacency_matrix)
            return self.graph_type(adjacency_matrix=adjacency_matrix,
                                   vertex_colors=np.zeros(self._graph_size, self._dtype))
        indices = np.nonzero(degrees)

        index = np.argwhere(indices == idx)[0][1]
        new_indices = np.delete(indices, index)

        length = len(new_indices)
        max_degree = np.min([max_degree, length])

        while max_degree > 0:
            choice = self._rnd_state.choice(new_indices, size=max_degree, replace=False)
            for k in choice:
                adjacency_matrix[idx][k], adjacency_matrix[k][idx] = 1, 1
                degrees[k] = degrees[k] - 1
            degrees[idx] = 0
            _amax = np.amax(degrees)
            arg_maxes = np.argwhere(degrees == _amax).flatten()
            idx = arg_maxes[self._rnd_state.randint(len(arg_maxes))]
            max_degree = degrees[idx]
            if max_degree == 0:
                break
            indices = np.nonzero(degrees)
            index = np.argwhere(indices == idx)[0][1]
            new_indices = np.delete(indices, index)
            length = len(new_indices)
            max_degree = np.min([max_degree, length])
        self._update_heatmap(adjacency_matrix)
        return self.graph_type(adjacency_matrix=adjacency_matrix, vertex_colors=np.zeros(self._graph_size, self._dtype))

    def __str__(self):
        """str: string representation of self"""
        return self.__class__.__name__ + "(delta={}, seed={})".format(self._max_delta, self._seed)

    def _get_max_delta(self):
        """int: returns maximum degree that each vertex of a generated graph instance is allowed to have"""
        return self._max_delta

    def _update_heatmap(self, matrix):
        """updates the heatmap of the generator"""
        _heatmap = self._heatmap * self._count + matrix
        self._count += 1
        self._heatmap = _heatmap / self._count

    def _get_heatmap(self):
        """np.ndarray: the heatmap of the generator"""
        return self._heatmap

    def save_heatmap(self, filepath, min_value=0.0, max_value=1.0):
        """Plots the heatmap and stores it at filepath.

        Attributes:
            filepath (str): filepath of plot to be stored at
            min_value (float): minimum value for heatmap (coldest value = blue)
            max_value (float): maximum value for heatmap (warmest value = yellow)

        Raises:
            DirectoryNotFoundError: If filepath can't be saved, since parent directories (the path) doesn't exist
        """
        if not os.path.exists(os.path.dirname(filepath)):
            raise DirectoryNotFoundError("directory '{}' does not exist".format(os.path.dirname(filepath)))
        plt.imshow(self.heatmap, cmap='viridis', vmin=min_value, vmax=max_value)
        plt.colorbar()
        plt.savefig(filepath)
        plt.close()

    max_delta = property(fset=None, fget=_get_max_delta, fdel=None)
    heatmap = property(fset=None, fget=_get_heatmap, fdel=None)



def memoize(func):
    """A function wrapper for a time/space trade-off."""
    table = dict()                      # function specific memoize table
    def wrappingfunction(*args):
        if args not in table:           # args tuple hasn't been seen yet
            table[args] = func(*args)   # envoke func call and store value
        return table[args]              # return stored value
    return wrappingfunction             # return wrappee


@memoize
def s2(n, k):
    """Calculates the Stirling number of the second kind."""
    if n == 0 or n != 0 and n == k:
        return 1
    if k == 0 or n < k:
        return 0
    return k * s2(n-1, k) + s2(n-1, k-1)


def str_iterable(iterable, depth=0, delimiter='    ', markdown_newline=False):
    """a helper function converting nested, iterable containers to str representation with visual indents for logging"""
    msg = "" if markdown_newline is False else "    "
    # no visual indent for len==0 iterables
    if len(iterable) == 0:
        if isinstance(iterable, list):
            return "[]"
        if isinstance(iterable, tuple):
            return "()"
        if isinstance(iterable, dict):
            return "{}"
    # do not nest lists or tuples with visual indents
    if isinstance(iterable, (list, tuple)):
        msg += "[\n"
        for value in iterable:
            msg += delimiter*(depth+1)
            if isinstance(value, (list, tuple, dict)):
                msg += "{},\n".format(str_iterable(value, depth+1))
            else:
                msg += "{},\n".format(str(value))
        msg += delimiter * depth + "]"
        return msg
    # nested visual indents for dictionaries
    if isinstance(iterable, dict):
        msg += "{\n"
        for key, value in iterable.items():
            msg += "{}{}: ".format(delimiter*(depth+1), str(key))
            if isinstance(value, (list, tuple, dict)):
                msg += "{},\n".format(str_iterable(value, depth+1))
            else:
                msg += "{},\n".format(str(value))
        msg += delimiter * depth + "}"
        return msg



from .math import s2
from .memory import Memory, ReplayMemory, PrioritizedReplayMemory
from .tensorboard import tensorboard_scalar, tensorboard_histo, tensorboard_array, tensorboard_text
from .load import load_hyperparameters
from .format import str_iterable
from .graph import Graph, NumpyGraph, EdgeListGraph, AdjacencyListGraph



import numpy as np

from .seqgen import SequenceGenerator


class RandomSequenceGenerator(SequenceGenerator):
    """A SequenceGenerator that generates sequence_number many sequences of length sequence_size. Pages are picked
    equally distributed from the interval [min_page, max_page].

    Args:
        sequence_size (int): target size of each sequence
        sequence_number (int): number of sequences to generate till generator depletes
        min_page (int): minimum element within each generated sequence
        max_page (int): maximum element within each generated sequence
        dtype (type): type of each element within sequence
        seed (int): seed for random number generation

    Attributes:
        _rnd_state (np.random.RandomState): random number generator for self with own local seed
        _seed (int or None): seed _rnd_state was initialized with
    """
    # noinspection PyUnusedLocal
    def __init__(self,
                 sequence_size,
                 sequence_number,
                 min_page,
                 max_page,
                 dtype=np.int32,
                 seed=None,
                 **kwargs):
        super(RandomSequenceGenerator, self).__init__(
            sequence_size=sequence_size,
            sequence_number=sequence_number,
            min_page=min_page,
            max_page=max_page,
            dtype=dtype
        )
        self._rnd_state = np.random.RandomState(seed=seed)
        self._seed = seed
        self._logger.info("initialized {}".format(self))

    def _generate_instance(self):
        """Create and return a newly generated, random, equally distributed sequence

        Returns:
            np.ndarray: equally distributed sequence with shape=(_sequence_size,) and dtype=_dtype
        """
        return self._rnd_state.randint(
            low=self._min_page,
            high=self._max_page+1,
            size=self._sequence_size,
            dtype=self._dtype
        )

    def __str__(self):
        """str: string representation of self"""
        return self.__class__.__name__ + "(seed={})".format(self._seed)


class RandomDistributionSequenceGenerator(SequenceGenerator):
    """A SequenceGenerator that generates sequence_number many sequences of length sequence_size.

    A probability_distribution is initially rolled with probability_distribution = normal(sigma, mu) of shape
    (num_pages,). afterwards softmax is applied on probability_distribution. Pages are picked from the interval
    [min_page, max_page] according to the softmaxed probability_distribution.

    Args:
        sequence_size (int): target size of each sequence
        sequence_number (int): number of sequences to generate till generator depletes
        min_page (int): minimum element within each generated sequence
        max_page (int): maximum element within each generated sequence
        sigma (float): sigma for normal(sigma, mu) probability distribution
        mu (float): mu for normal(sigma, mu) probability distribution
        dtype (type): type of each element within sequence
        seed (int): seed for random number generation

    Attributes:
        _rnd_state (np.random.RandomState): random number generator for self with own local seed
        _seed (int or None): seed _rnd_state was initialized with
        _sigma (float): sigma for normal(sigma, mu) probability distribution
        _mu (float): mu for normal(sigma, mu) probability distribution
        _probability_distr (np.npdarray, shape=(_num_pages,), dtype=np.float32): probability distribution from softmaxed normal sample
    """
    # noinspection PyUnusedLocal
    def __init__(self,
                 sequence_size,
                 sequence_number,
                 min_page,
                 max_page,
                 sigma=1.0,
                 mu=0.0,
                 dtype=np.int32,
                 seed=None,
                 **kwargs):
        super(RandomDistributionSequenceGenerator, self).__init__(
            sequence_size=sequence_size,
            sequence_number=sequence_number,
            min_page=min_page,
            max_page=max_page,
            dtype=dtype
        )
        self._rnd_state = np.random.RandomState(seed=seed)
        self._seed = seed
        self._sigma = sigma
        self._mu = mu
        self._generate_probability_distr()
        self._logger.info("initialized {} with probability_distribution='{}'".format(self, self._probability_distr))

    def _generate_probability_distr(self):
        """rolls the distribution normal(sigma, mu) and applies a softmax to it"""
        # roll distribution
        self._probability_distr = self._sigma * self._rnd_state.randn(self._num_pages) + self._mu
        # apply softmax to distribution
        self._probability_distr = np.exp(self._probability_distr) / np.sum(np.exp(self._probability_distr))

    def _generate_instance(self):
        """Create and return a newly random, from normal(sigma, mu)-distributed softmaxed sample generated, sequence

        Returns:
            np.ndarray: sequence with shape=(_sequence_size,) and dtype=_dtype drawn from _probability_distr
        """
        return self._rnd_state.choice(a=self._pages, size=self._sequence_size, p=self._probability_distr)

    def _get_probability_distr(self):
        """np.ndarray: softmaxed probability distribution with shape=(_num_pages,) and dtype=np.float32"""
        return self._probability_distr

    def _set_probability_distr(self, value):
        assert isinstance(value, np.ndarray), "probability_distribution must be of type='{}'".format(np.ndarray)
        assert value.shape == (self._num_pages,), "probability_distribution must be of shape='{}' but " \
            "received shape='{}'".format((self._num_pages,), value.shape)
        assert np.sum(value) == 1.0, "probability_distribution's sum must equal 1.0"
        self._probability_distr = value

    def __str__(self):
        """str: string representation of self"""
        return self.__class__.__name__ + "(seed={}, \u03C3={}, \u03BC={})".format(self._seed, self._sigma, self._mu)

    probability_distribution = property(fset=_set_probability_distr, fget=_get_probability_distr, fdel=None)



import numpy as np
from abc import abstractmethod
from collections.abc import Iterable, Iterator

from .sumtree import SumTree


class Memory(Iterator):
    """Abstract Replay Memory base class for Agents to store experiences in and draw from.

    Each Memory implementation must be agnostic with regards to its experiences; hence, return types of Memory are
    referred to as type(experience).

    Args:
        memory_size (int): number of experiences to store

    Attributes:
        _memory_size(int): number of experiences to store

    Raises:
        NotImplementedError: if abstract method is to be invoked
    """
    def __init__(self, memory_size):
        self._memory_size = memory_size
        if type(self) == Memory:
            raise NotImplementedError("{} must be subclassed.".format(self.__class__.__name__))

    @abstractmethod
    def __getitem__(self, item):
        """get experiences(s) indexed with item

        Args:
            item (number, slice, Iterable): a selector to index experience with in self

        Returns:
             type(experience): if isinstance(item, number)
             tuple of type(experience): if isinstance(item, (Iterable, slice))

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    def __iter__(self):
        """generator: generator to iterate with over self"""
        return self.__next__()

    def __next__(self):
        """type(experience): iterate over len, yield experience at index"""
        for index in range(len(self)):
            yield self[index]

    @abstractmethod
    def __len__(self):
        """int: number of currently stored experiences"""
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def clear(self):
        """clears Memory and resets all necessary data structures"""
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def sample(self, sample_size):
        """Draw experiences of size sample_size.

        Args:
            sample_size (int): sample size to be drawn

        Returns:
            tuple of type(experience): tuple of drawn experiences with length sample_size
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def store(self, experience):
        """add experience to Memory"""
        raise NotImplementedError("Must be implemented in subclass.")


class ReplayMemory(Memory):
    """A light-weight ring buffer for containing experiences. Experiences are drawn uniformly.

    Args:
        memory_size (int): number of experiences to store
        seed (int or None): seed for random choice

    Attributes:
        _data (list): internal container for experiences to be stored
        _data_cursor (int): cursor for ring-buffer implementation
        _rnd_state (np.random.RandomState): random number generator for self with own local seed
        _seed (int or None): seed _rnd_state was initialized with
    """
    def __init__(self, memory_size, seed=None):
        super().__init__(memory_size=memory_size)
        self._data = list()
        self._data_cursor = 0
        self._seed = seed
        self._rnd_state = np.random.RandomState(seed=seed)

    def __setitem__(self, index, value):
        """set value/s at index/indices"""
        self._data[index] = value

    def __getitem__(self, item):
        """Retrieve experience(s) stored in ReplayMemory indexed with item.

        Args:
            item (number, slice, Iterable): a selector to index experience with in self

        Returns:
             type(experience): if isinstance(item, number)
             tuple of type(experience): if isinstance(item, (Iterable, slice))

        Raises:
            TypeError: item selection not defined with type of item
        """
        if isinstance(item, Iterable):
            return tuple(map(self._data.__getitem__, item))
        if isinstance(item, slice):
            return tuple(self._data[item])
        if isinstance(item, (int, float, np.number)):
            return self._data[item]
        raise TypeError("{} item is of unexpected type='{}': can't perform get operation"
                        .format(self.__class__.__name__, type(item)))

    def __len__(self):
        """int: len of self"""
        return len(self._data)

    def store(self, experience):
        """Store experience in Memory. If _memory_size is not reached yet, append experience; else store at _data_cursor
        and move _data_cursor.

        Args:
            experience (type(experience)): experience to store
        """
        if len(self) < self._memory_size:
            self._data.append(experience)
        else:
            self[self._data_cursor] = experience
            self._data_cursor = (self._data_cursor + 1) % self._memory_size

    def clear(self):
        """Clear memory."""
        del self._data[:]
        self._data_cursor = 0

    def sample(self, sample_size):
        """Uniformly draw sample of size sample_size from Memory.

        Args:
            sample_size (int): sample size to be drawn

        Returns:
            tuple of type(experience): tuple of drawn experiences with length sample_size
        """
        rnd_idx = self._rnd_state.randint(low=0, high=len(self), size=sample_size, dtype=np.int32)
        return self[rnd_idx]


class PrioritizedReplayMemory(Memory):
    """An implementation of Prioritized Replay Memory according to Schaul et. al.: https://arxiv.org/pdf/1511.05952.pdf

    Prioritizes important experiences to be picked more frequently based on td_errors from a QAgent. This implementation
    uses a binary indexed tree as an indexing structure to sample 'more valuable' experiences (commonly referred to as
    SumTree or Fenwick tree). For stability reasons td_errors have to be normalized via td_normalization to fall within
    [0.0, 1.0]. Priority updates within the index structure can result in small delta updates, which can not be
    propagated upwards throughout the entire tree due to floating point precision limitations; the SumTree
    implementation at hand checks for "failing" delta updates and auto-corrects "too small delta changes" to the next
    applicable machine epsilon. If the auto-correct exceeds a relative change of max_relative_change, a warning is
    logged.

    Args:
        memory_size (int): number of experiences to store
        alpha (float): defines how much prioritization is used (alpha = 0.0 correspond to the uniform case)
        beta (float): defines how much importance sampling is used (linearly increased for each global step)
        beta_delta (float): defines by how much beta is increased after each global step
        eps (float): smallest possible priority to clip 0.0
        td_normalization (number): value by which td_errors is divided to normalize td_error in [0.0, 1.0]
        max_relative_change (number): maximum relative change of SumTree's auto-correction for numerical stability
        tree_dtype (type): dtype for floating point precision used in SumTree
        seed (int or None): seed for random choice

    Attributes:
        _alpha (float): defines how much prioritization is used (alpha = 0.0 correspond to the uniform case)
        _beta (float): defines how much importance sampling is used (linearly increased for each global step)
        _beta_delta (float): defines by how much beta is increased after each global step
        _tree (utils.SumTree): a Fenwick tree instance that indexes _memory_size many experiences
        _eps (float): smallest possible priority to clip 0.0
        _td_normalization (number): value by which td_errors is divided to normalize td_error in [0.0, 1.0]
        _rnd_state (np.random.RandomState): random number generator for self with own local seed
        _seed (int or None): seed for random choice
    """
    def __init__(self,
                 memory_size,
                 alpha,
                 beta,
                 beta_delta,
                 eps,
                 td_normalization,
                 max_relative_change=1e3,
                 tree_dtype=np.float64,
                 seed=None,
                 **kwargs):
        super().__init__(memory_size=memory_size)
        self.alpha = alpha
        self.beta = beta
        self.beta_delta = beta_delta
        self._tree = SumTree(memory_size=self._memory_size, max_relative_change=max_relative_change, dtype=tree_dtype)
        self._eps = eps
        self._td_normalization = td_normalization
        self._seed = seed
        self._rnd_state = np.random.RandomState(seed=self._seed)

    def __len__(self):
        """int: number of experiences stored"""
        return self._tree.data_size

    def __getitem__(self, item):
        """Retrieve experience(s) stored in ReplayMemory indexed with item.

        Args:
            item (number, slice, Iterable): a selector to index experience with in _tree

        Returns:
             tuple of type(experience):
        """
        return self._tree[item]

    def clear(self):
        """Clear memory."""
        self._tree.clear()

    # noinspection PyMethodOverriding
    def sample(self, sample_size):
        """Draw sample of size sample_size from Memory based on priorities.

        Args:
            sample_size (int): sample size to be drawn

        Returns:
            tuple(tuple(int), tuple(float), list(type(experience))): First entry depicts tree index at which each
                experience was drawn from (reused for update after batch is trained on). Second entry depicts priority
                value of experience. Last entry is the experience itself.
        """
        _total_priority = self._tree.total_priority
        _alpha_priority = np.sum(np.power(self._tree.priorities, self.alpha))
        _segment_num = int(_total_priority // sample_size)
        _segment_values = [self._rnd_state.uniform(_segment_num*i, _segment_num*(i + 1)) for i in range(sample_size)]
        _tree_indices, _priorities, _experiences = zip(*self[_segment_values])
        _sampled_weights = [(self._memory_size * (p ** self.alpha) / _alpha_priority) ** (-self.beta) for p in _priorities]
        _max_sample_weight = max(_sampled_weights)
        _sampled_weights[:] = [s / _max_sample_weight for s in _sampled_weights]
        return _tree_indices, _sampled_weights, _experiences

    def store(self, experience):
        """Store experience in Memory. Each new experience is initialized with max_priority to ensure it's trained upon
        at least once.

        Args:
            experience type(experience): experience to be stored
        """
        _max_priority = self._tree.max_priority
        _max_priority = _max_priority if _max_priority != 0.0 else 1.0
        self._tree.add(priority=_max_priority, data_entry=experience)

    def update(self, tree_indices, td_errors):
        """Updates memory after last sample (batch) is processed.

        Args:
            tree_indices (tuple of int): tree index at which experience was drawn from prior sample() invocation
            td_errors (np.ndarray of float): td_error for each experience from prior sample() invocation
        """
        td_errors = td_errors / self._td_normalization              # normalize td_error to be in [0.0, 1.0]
        td_errors[td_errors == 0.0] += self._eps                    # clip to be in [self._eps, 1.0 + self._eps]
        td_errors = np.minimum(td_errors, 1.0)                      # clip to be in [self._eps, 1.0]
        for tree_index, td_error in zip(tree_indices, td_errors):   # update tree priorities
            self._tree.update(tree_index, td_error)

    def increment_beta(self):
        """increment beta by beta_delta clipped at 1.0"""
        self.beta = min(self.beta + self.beta_delta, 1.0)

    def _set_alpha(self, value):
        assert 0.0 <= value <= 1.0, "alpha must be in interval [0.0, 1.0] but received value='{}'instead".format(value)
        self._alpha = value

    def _set_beta(self, value):
        assert 0.0 <= value <= 1.0, "beta must be in interval [0.0, 1.0] but received value='{}'instead".format(value)
        self._beta = value

    def _set_beta_delta(self, value):
        assert 0.0 <= value <= 1.0, "beta_delta must be in interval [0.0, 1.0] but received value='{}'instead"\
            .format(value)
        self._beta_delta = value

    def _get_alpha(self):
        """float: defines how much prioritization is used (alpha = 0.0 correspond to the uniform case)"""
        return self._alpha

    def _get_beta(self):
        """float: defines how much importance sampling is used (linearly increased for each global step)"""
        return self._beta

    def _get_beta_delta(self):
        """float: defines by how much beta is increased after each global step"""
        return self._beta_delta

    alpha = property(fset=_set_alpha, fget=_get_alpha, fdel=None)
    beta = property(fset=_set_beta, fget=_get_beta, fdel=None)
    beta_delta = property(fset=_set_beta_delta, fget=_get_beta_delta, fdel=None)


import logging
from abc import abstractmethod, ABCMeta

import numpy as np


class Environment(metaclass=ABCMeta):
    """Abstract base class for Environments.

    Attributes:
        _logger (logging.Logger): logger instance for each subclassed instance
    """
    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self._logger.info("initializing {}".format(self.__class__.__name__))

    @abstractmethod
    def _get_action_dim(self):
        """Environment returns number of all actions.

        Returns:
            int: number of actions

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def _get_state_dim(self):
        """Environment returns length of state vector.

        Returns:
            int: length of state vector

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def get_state(self, vectorized_state=False, dtype=np.int32):
        """Environment returns state.

        Attributes:
            vectorized_state (bool): depicts if return type should be vectorized (np.ndarray) or non-vectorized (dict)

        Returns:
            np.ndarray or dict: state of environment

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def calculate_state_complexity(self):
        """Environment calculates its state complexity.

        Returns:
            int: cardinality of state space

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def step(self, action, vectorized_state=True):
        """Environment performs action and transition into next state.

        Attributes:
            action (int): action to take in current state
            vectorized_state (bool): depicts if return type should be vectorized (np.ndarray) or non-vectorized (dict)

        Returns:
            tuple: first element depicts the reward for action taken as a float number; second element is the state
                either as a np.ndarray or dict depending on vectorized_state (True for np.ndarray, False for dict)

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def reset(self):
        """Environment resets and sets its internal structures to default values.

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def _get_instance(self):
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def _set_instance(self, value):
        raise NotImplementedError("Must be implemented in subclass.")

    instance = property(fset=_set_instance, fget=_get_instance, fdel=None)


class PagingEnvironment(Environment):
    """Abstract base class for pyloa.agent.PagingAgent agents to play on.

    Args:
        sequence_size (int): number of pages per sequence
        cache_size (int): number of pages in cache
        num_pages (int): number of different pages
        include_request (bool): depicts whether or not request (next page) is included within state
        include_pagefault (bool): depicts whether or not is_pagefault is included within state
        include_cache (bool): depicts whether or not _cache is included within state
        include_time_in_cache (bool): depicts whether or not _time_im_cache is included within state
        include_time_since_touched (bool): depicts whether or not _time_since_touched is included within state
        include_histogram (bool): depicts whether or not _histogram is included within state

    Attributes:
         _sequence_size (int): number of pages per sequence
         _cache_size (int): number of pages in cache
         _num_pages (int): number of different pages
         _num_steps (int): counter for current steps taken by agent so far
         _sequence (np.ndarray, shape=(_sequence_size,)): sequence of pages
         _cache (np.ndarray, shape=(_cache_sitze,), dtype=np.int32): pages in cache
         _time_in_cache (np.ndarray, shape=(_cache_sitze,), dtype=np.int32): time since page at index is in cache
         _histogram (np.ndarray, shape=(_cache_size,), dtype=np.int32): number of total requests for page in cache
         _time_since_touched (np.ndarray, shape=(_num_pages,), dtype=np.int32): time since page at index was requested
         _include_request (bool): depicts whether or not request (next page) is included within state
         _include_pagefault (bool): depicts whether or not is_pagefault is included within state
         _include_cache (bool): depicts whether or not _cache is included within state
         _include_time_in_cache (bool): depicts whether or not _time_im_cache is included within state
         _include_time_since_touched (bool): depicts whether or not _time_since_touched is included within state
         _include_histogram (bool): depicts whether or not _histogram is included within state

    Raises:
        NotImplementedError: if abstract class is to be instanced
    """
    def __init__(self,
                 sequence_size,
                 cache_size,
                 num_pages,
                 include_request=True,
                 include_pagefault=False,
                 include_cache=False,
                 include_time_in_cache=False,
                 include_time_since_touched=False,
                 include_histogram=False,
                 **kwargs):
        super().__init__()
        # params
        self._sequence_size = sequence_size
        self._cache_size = cache_size
        self._num_pages = num_pages
        self._num_steps = 0
        # structures
        self._sequence = np.zeros(self._sequence_size, dtype=np.int32)
        self._cache = np.zeros(self._cache_size, dtype=np.int32)
        self._time_in_cache = np.zeros(self._cache_size, dtype=np.int32)
        self._histogram = np.zeros(self._cache_size, dtype=np.int32)
        self._time_since_touched = np.zeros(self._cache_size, dtype=np.int32)
        self._default_cache = np.arange(start=1, stop=self._cache_size + 1)
        self._default_time_in_cache = np.flip(np.arange(start=1, stop=self._cache_size + 1))
        self._default_histogram = np.zeros(self._cache_size, dtype=np.int32)
        self._default_time_since_touched = np.flip(np.arange(start=1, stop=self._cache_size + 1))
        # state config
        self._include_request = include_request
        self._include_pagefault = include_pagefault
        self._include_cache = include_cache
        self._include_time_in_cache = include_time_in_cache
        self._include_time_since_touched = include_time_since_touched
        self._include_histogram = include_histogram

    def _get_action_dim(self):
        """int: number of actions"""
        return self._cache_size + 1

    def _get_cache(self):
        """np.ndarray: current cache of environment"""
        return self._cache

    def _get_cache_size(self):
        """int: max. number of pages in cache"""
        return self._cache_size

    def _get_histogram(self):
        """np.ndarray: current histogram of environment"""
        return self._histogram

    def _get_num_pages(self):
        """int: number of different pages (excludes '0' as placeholder for default page)"""
        return self._num_pages

    def _get_num_steps(self):
        """int: current number of actions/steps played on _sequence"""
        return self._num_steps

    def _get_request(self):
        """int: current page that is requested and has to be placed in cache"""
        return self._sequence[self.num_steps]

    def _get_time_in_cache(self):
        """np.ndarray: current _time_in_cache of environment"""
        return self._time_in_cache

    def _get_time_since_touched(self):
        """np.ndarray: current _time_since_touched of environment"""
        return self._time_since_touched

    def _get_instance(self):
        """np.ndarray: current sequence that agent plays on"""
        return self._sequence

    def _get_state_dim(self):
        """int: length of state vector"""
        return len(self.get_state(vectorized_state=True))

    def _get_used_indices(self):
        """np.ndarray: current indices in _cache that are beiung used (non-placeholder/default value)"""
        return np.where(self._cache != 0)[0]

    def _is_terminated(self):
        """bool: True, if environment is terminated (all requests for _sequence evaluated). False otherwise."""
        return self.num_steps == len(self._sequence) - 1

    def _is_pagefault(self):
        """bool: True, if requested page is not in _cache (page fault), else False."""
        return self.request not in self.cache

    def _set_instance(self, value):
        """Overrides currently stored _sequence with value without invoking reset()

        Args:
            value (np.ndarray): new _sequence of requests to store and play on

        Raises:
            AssertionError: if length of value does not equal _sequence_size or if value is not of type np.ndarray
        """
        assert isinstance(value, np.ndarray), "value must be of type='{}' but received type='{}' instead."\
            .format(np.ndarray, type(value))
        assert len(value) == self._sequence_size, "value must be of length='{}' but received length='{}' instead."\
            .format(self._sequence_size, len(value))
        self._sequence = value

    @abstractmethod
    def _calculate_cache_state_complexity(self):
        raise NotImplementedError("Must be implemented in subclass.")

    def _calculate_pagefault_state_complexity(self):
        """Calculates and returns the state complexity of page fault states, if page fault is included within
        environment's state representation; otherwise returns 1.

        Returns:
            int: 2, if page fault is included, else 1
        """
        return 2 if self._include_pagefault else 1

    def _calculate_request_state_complexity(self):
        """Calculates and returns the state complexity of request states, if request is included within environment's
        state representation; otherwise returns 1.

        Returns:
            int: _num_pages if request is included, else 1
        """
        return self.num_pages if self._include_request else 1

    @abstractmethod
    def _calculate_time_in_cache_state_complexity(self):
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def _calculate_time_since_touched_state_complexity(self):
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def _calculate_histogram_state_complexity(self):
        raise NotImplementedError("Must be implemented in subclass.")

    def calculate_state_complexity(self):
        """Calculates and returns the cardinality of state space.

        Returns:
            int: cardinality of state space
        """
        return self._calculate_request_state_complexity() * \
            self._calculate_pagefault_state_complexity() * \
            self._calculate_cache_state_complexity() * \
            self._calculate_time_in_cache_state_complexity() * \
            self._calculate_time_since_touched_state_complexity() * \
            self._calculate_histogram_state_complexity()

    def get_state(self, vectorized_state=False, vectorized_dtype=np.int32):
        """Returns the state of environment. The state is either vectorized (returns np.ndarray) or not (returns dict).
        The state comprises copies of each variable that has been defined to be included during environment creation.

        Args:
            vectorized_state: defines return type. A vectorized_state will be of type np.ndarray, otherwise dict
            vectorized_dtype: dtype of np.darray for vectorized_state

        Returns:
            np.ndarray or dict: state depicting the current observation (changing the returned state won't change the
                internal structures in environment)
        """
        if vectorized_state:
            _state = np.asarray([], dtype=vectorized_dtype)
            if self._include_request:
                _state = np.append(_state, np.asarray(self.request, dtype=vectorized_dtype))
            if self._include_pagefault:
                _state = np.append(_state, np.asarray([1 if self.is_pagefault else 0], dtype=vectorized_dtype))
            if self._include_cache:
                _state = np.append(_state, self.cache)
            if self._include_time_in_cache:
                _state = np.append(_state, self.time_in_cache)
            if self._include_time_since_touched:
                _state = np.append(_state, self.time_since_touched)
            if self._include_histogram:
                _state = np.append(_state, self.histogram)
        else:
            _state = dict()
            _state['current_page'] = self.request
            _state['is_pagefault'] = self.is_pagefault
            _state['cache'] = np.copy(self.cache)
            _state['time_in_cache'] = np.copy(self.time_in_cache)
            _state['time_since_touched'] = np.copy(self.time_since_touched)
            _state['histogram'] = np.copy(self.histogram)
            _state['num_steps'] = self.num_steps
        return _state

    def reset(self, cache=None, time_in_cache=None, time_since_touched=None, histogram=None):
        """Resets the environment to its initial state.

        Args:
            cache (collections.abc.Sequence): initial cache after reset
            time_in_cache (collections.abc.Sequence): initial time_in_cache after reset
            time_since_touched (collections.abc.Sequence): initial time_since_touched after reset
            histogram (collections.abc.Sequence): initial histogram after reset
        """
        self._num_steps = 0
        self._cache[:] = self._default_cache if cache is None else cache
        self._time_in_cache[:] = self._default_time_in_cache if time_in_cache is None else time_in_cache
        self._time_since_touched[:] = self._default_time_since_touched if time_since_touched is None else \
            time_since_touched
        self._histogram[:] = self._default_histogram if histogram is None else histogram

    def step(self, action, vectorized_state=True):
        """Environment computes the step for action, updates all internal structures and returns the next state and cost
        as a tuple.

        Args:
            action (int): depicts the action (index) that the agent wants to flush (or 0 for no-action)
            vectorized_state (bool): depicts the return type of the observation after step/action is taken

        Returns:
            tuple: first entry is the next state, second entry are the costs affiliated with last action in last state

        Raises:
            AssertionError: if action is not in interval [0, cache_size] and therefore invalid
        """
        # check bounds (0 means 'no action',  interval [1, cache_size] represents indices (offset by one))
        assert 0 <= action <= self._cache_size, "action must be in interval [0, {}] but received action='{}' instead"\
            .format(self._cache_size, action)
        used_indices = self.used_indices

        # no action (no flush, current_page expected by agent to exist in cache)
        if action == 0:
            # agent is right, page exists in cache -> update structures
            if not self.is_pagefault:
                idx = np.where(self._cache == self.request)[0][0]
                self._time_since_touched[used_indices] += 1
                self._time_since_touched[idx] = 0
                self._time_in_cache[used_indices] += 1
                self._histogram[idx] += 1
                cost = 0  # no costs, since agent correctly identified that current_page is in cache
            # agent is incorrect, page doesn't exist in cache
            else:
                self._time_since_touched[used_indices] += 1
                self._time_in_cache[used_indices] += 1
                cost = 1000  # high costs, since agent falsely expected current_page to be in cache
        # agent choose action to flush index to place current_page
        else:
            # agent tries to flush an index, even though current_page is already in cache
            if not self.is_pagefault:
                idx = np.where(self._cache == self.request)[0][0]
                self._time_since_touched[used_indices] += 1
                self._time_since_touched[idx] = 0
                self._time_in_cache[used_indices] += 1
                self._histogram[idx] += 1
                cost = 500
            # agent is right, page doesn't exist in cache -> flush selected action index
            else:
                idx = action - 1
                # agent decides to place current_page in empty cache slot -> no flush
                if 0 in self._cache and self._cache[idx] == 0:
                    cost = 0
                # agent decide to place current_page in a non-empty index, even though there is free space
                elif 0 in self._cache and self._cache[idx] != 0:
                    cost = 250
                # normal flush
                else:
                    cost = 10
                self._cache[idx] = self.request
                self._time_since_touched[used_indices] += 1
                self._time_since_touched[idx] = 0
                self._time_in_cache[used_indices] += 1
                self._time_in_cache[idx] = 1
                self._histogram[idx] = 1
        # inc step
        self._num_steps += 1
        return self.get_state(vectorized_state=vectorized_state), cost

    action_dim = property(fset=None, fget=_get_action_dim, fdel=None)
    cache = property(fset=None, fget=_get_cache, fdel=None)
    cache_size = property(fset=None, fget=_get_cache_size, fdel=None)
    histogram = property(fset=None, fget=_get_histogram, fdel=None)
    instance = property(fset=_set_instance, fget=_get_instance, fdel=None)
    is_pagefault = property( fset=None, fget=_is_pagefault, fdel=None)
    is_terminated = property(fset=None, fget=_is_terminated, fdel=None)
    num_pages = property(fset=None, fget=_get_num_pages, fdel=None)
    num_steps = property(fset=None, fget=_get_num_steps, fdel=None)
    request = property(fset=None, fget=_get_request, fdel=None)
    state_dim = property(fset=None, fget=_get_state_dim, fdel=None)
    used_indices = property(fset=None, fget=_get_used_indices, fdel=None)
    time_in_cache = property(fset=None, fget=_get_time_in_cache, fdel=None)
    time_since_touched = property(fset=None, fget=_get_time_since_touched, fdel=None)



from abc import abstractmethod

import numpy as np

from .agent import Agent
from .qagent import QAgent
from .rlagent import RLAgent
from pyloa.utils.tensorboard import tensorboard_scalar, tensorboard_array, tensorboard_histo


class PagingAgent(Agent):
    """Abstract base class for Agents that play on pyloa.environment.PagingEnvironment sequences.

    Args:
        state_dim (int): defines input shape=(_state_dim,) of input states
        action_dim (int): defines the number of different actions the agent can take
        name (str): depicts a name for the agent
    """
    def __init__(self, state_dim, action_dim, name=None, **kwargs):
        super().__init__(state_dim=state_dim, action_dim=action_dim, name=name)

    # noinspection PyUnresolvedReferences
    # noinspection PyAttributeOutsideInit
    def play_instance(self, environment, is_train, episode, writer, invert_cost=True):
        """Agent plays on environment for one episode (on one sequence).

        Args:
            environment (pyloa.environment.PagingEnvironment): environment to play on
            episode (int): number depicting how many episodes agent played so far
            is_train (bool): depicts if agent should train (True) or just evaluate (False)
            writer (tf.summary.FileWriter): tensorboard logging for playing a sequence
            invert_cost (bool): depicts if costs are inverted after calculation (cost = negative reward)

        Returns:
            tuple: agent will return its calculated metrics of this episode as follows (steps, cost, page_faults)
                    steps (int): depicts the number of actions till the environment terminated
                    cost (int): total cost accumulated during episode
                    page_faults (int): depicts the number of page faults the agent had during episode

        Raises:
            RuntimeError: if agent is invoked with is_train=True, but agent is not a subclass of RLAgent
        """
        _is_trainable = isinstance(self, RLAgent)               # has epsilon and train
        _has_memory = isinstance(self, QAgent)                  # has epsilon, memory and train
        # temp vars
        _temp_cost = 0
        _temp_num_page_fault = 0
        _temp_actions = []
        _temp_losses = []
        _temp_epsilon = 0.0
        _temp_num_cognition_fault = 0

        # sanity check to make sure only trainable agents can be trained
        if is_train and not _is_trainable:
            raise RuntimeError("{} called to learn but agent isn't trainable!".format(self.__class__.__name__))

        # change epsilon if agent is trainable but isn't training right now for eval
        if _is_trainable and not is_train:
            _temp_epsilon = self.epsilon
            self.epsilon = 1.0

        # play environment till it terminates
        while not environment.is_terminated:
            # get current state (trainable agents expect vecotrized np.ndarray states)
            state = environment.get_state(vectorized_state=_is_trainable)

            # get next predicted action
            action = self.act(state=state)

            # update page fault metrics: a page fault occurred
            _temp_num_page_fault += environment.is_pagefault
            # update cognition fault metrics: agent took action even though he shouldn't have OR
            # agent took no action even though he should have)
            _temp_num_cognition_fault += environment.is_pagefault ^ (action > 0)

            # take action, invert cost
            next_state, cost = environment.step(action=action, vectorized_state=_is_trainable)
            cost = -cost if invert_cost else cost

            # update action_histogram, cost temp metrics
            _temp_actions.append(action)
            _temp_cost += abs(cost)

            # store action if agent is training and has memory
            if _has_memory and is_train:
                self.store(state=state, action=action, reward=cost, next_state=next_state)

            # train episodic non-QAgents, append loss values
            if not _has_memory and _is_trainable and is_train:
                _loss = self.train(state=state, action=action, reward=cost, next_state=next_state)
                _temp_losses.append(_loss)

        # train replay memory QAgents
        if _has_memory and is_train:
            _temp_losses = self.train(train_writer=writer)

        # adapt epsilon
        if _is_trainable:
            if is_train:
                self.increment_epsilon()        # increment epsilon if agent is trainable and is training
            else:
                self.epsilon = _temp_epsilon    # change epsilon back if agent is trainable but isn't training right now

        # store metrics
        if writer is not None:
            tensorboard_scalar(writer, 'total_cost/' + str(self), _temp_cost, episode)
            tensorboard_scalar(writer, 'avg_cost/' + str(self), _temp_cost / environment.num_steps, episode)
            tensorboard_scalar(writer, 'miss_ratio/' + str(self), _temp_num_page_fault / environment.num_steps, episode)
            tensorboard_scalar(writer, 'cognition_fault_ratio/' + str(self), _temp_num_cognition_fault /
                               environment.num_steps, episode)
            tensorboard_scalar(writer, 'error_ratio/' + str(self), (_temp_num_page_fault + _temp_num_cognition_fault) /
                               environment.num_steps, episode)
            tensorboard_histo(writer, 'actions/' + str(self), _temp_actions, episode)
            if is_train:
                tensorboard_array(writer, 'episode/loss', _temp_losses, episode)
                tensorboard_scalar(writer, 'episode/epsilon', self.epsilon, episode)

        # return performance
        return environment.num_steps, _temp_cost, _temp_num_page_fault

    @abstractmethod
    def act(self, state):
        """Agent decides on an action based on information given with state.

        Args:
            state (np.ndarray with shape=(_state_dim,) or dict): input state

        Returns:
            int: action to be taken

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")


class FIFOAgent(PagingAgent):
    """Deterministic First-In-First-Out agent; on-line, conservative, theoretical competitive ratio = cache_size"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def act(self, state):
        """Agent evicts oldest page in cache.

        Args:
            state (dict): input state

        Returns:
            int: action to be taken
        """
        if not state['is_pagefault']:                   # if the current state is not a page fault
            return 0                                    # don't evict a slot (0 = no action)
        return np.argmax(state['time_in_cache']) + 1    # evict oldest page with action 'index+1'


class LIFOAgent(PagingAgent):
    """Deterministic Last-In-First-Out agent; on-line, non-conservative, theoretical competitive ratio is unbound"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def act(self, state):
        """Agent evicts youngest page in cache,

        Args:
            state (dict): input state

        Returns:
            int: action to be taken
        """
        if not state['is_pagefault']:                   # if the current state is not a page fault
            return 0                                    # don't evict a slot (0 = no action)
        return np.argmin(state['time_in_cache']) + 1    # evict youngest page with action 'index+1'


class MostFrequentlyUsedAgent(PagingAgent):
    """Deterministic Most-Frequently-Used agent; on-line, non-conservative, theo. competitive ratio = cache_size"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def act(self, state):
        """Agent evicts most-accessed page in cache,

        Args:
            state (dict): input state

        Returns:
            int: action to be taken
        """
        if not state['is_pagefault']:                   # if the current state is not a page fault
            return 0                                    # don't evict a slot (0 = no action)
        return np.argmax(state['histogram']) + 1        # evict most-accessed page with action 'index+1'


class LeastFrequentlyUsedAgent(PagingAgent):
    """Deterministic Least-Frequently-Used agent; on-line, non-conservative, theoretical competitive ratio is unbound"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def act(self, state):
        """Agent evicts least-accessed page in cache.

        Args:
            state (dict): input state

        Returns:
            int: action to be taken
        """
        if not state['is_pagefault']:                   # if the current state is not a page fault
            return 0                                    # don't evict a slot (0 = no action)
        return np.argmin(state['histogram']) + 1        # evict least-accessed oage with action 'index+1'


class MostRecentlyUsedAgent(PagingAgent):
    """Deterministic Most-Recently-Used agent; on-line, non-conservative, theoretical competitive ratio is unbound"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def act(self, state):
        """Agent evicts most-recently-accessed page in cache.

        Args:
            state (dict): input state

        Returns:
            int: action to be taken
        """
        if not state['is_pagefault']:                       # if the current state is not a page fault
            return 0                                        # don't evict a slot (0 = no action)
        return np.argmin(state['time_since_touched']) + 1   # evict most-recently-accessed page with action 'index+1'


class LeastRecentlyUsedAgent(PagingAgent):
    """Deterministic Least-Recently-Used agent; on-line, non-conservative, theoretical competitive ratio = cache_size"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def act(self, state):
        """Agent evicts least-recently-accessed page in cache.

        Args:
            state (dict): input state

        Returns:
            int: action to be taken
        """
        if not state['is_pagefault']:                       # if the current state is not a page fault
            return 0                                        # don't evict a slot (0 = no action)
        return np.argmax(state['time_since_touched']) + 1   # evict least-recently-accessed page with action 'index+1'


class RandomAgent(PagingAgent):
    """Non-Deterministic randomized agent; on-line, non-conservative, theoretical competitive ratio is unclear

    Args:
        seed (int): sets np.random.seed for reproducible randomization

    Attributes:
        _rnd_state (np.random.RandomState): random number generator for self with own local seed
        _seed (int):  sets np.random.seed for reproducible randomization
    """
    # noinspection PyUnusedLocal
    def __init__(self, seed=None, **kwargs):
        self._seed = seed
        self._rnd_state = np.random.RandomState(seed=seed)
        super().__init__(**kwargs)

    def act(self, state):
        """Agent evicts least-recently-accessed page in cache,

        Args:
            state (dict): input state

        Returns:
            int: action to be taken
        """
        if not state['is_pagefault']:                                   # if the current state is not a page fault
            return 0                                                    # don't evict a slot (0 = no action)
        return self._rnd_state.randint(low=0, high=len(state['cache'])) + 1   # evict randomly chosen 'index+1'

    def __str__(self):
        """str: string representation of self"""
        return self.__class__.__name__ + "(seed={})".format(self._seed)


class BeladysOptimalAgent(PagingAgent):
    """Deterministic optimal agent; off-line, conservative, theoretical competitive ratio = 1.0

    Belady's Algorithm is the optimal, off-line algorithm for the paging problem. The agent overloads super's
    play_instance to calculate an off-line solution via _calc_opt - inspecting the sequence that environment is going to
    iterate over (looking into the future). Within _calc_opt the agent fills its list actions with the optimal action to
    take at each time step. Afterwards, the agent invokes super's play_instance, whereas each function call of act
    utilizes the off-line information gathered from _calc_opt.

    Args:
        state_dim (int): defines input shape=(_state_dim,) of input states
        action_dim (int): defines the number of different actions the agent can take

    Attributes:
        _current_time_step (int): time step that increases after each act and resets for a new play_instance call
        _actions (list of int): optimal action to take at each time step
    """
    # noinspection PyUnusedLocal
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._current_time_step = 0
        self._actions = list()

    def _calc_opt(self, sequence, cache):
        """Calculates optimal action at each time step to play sequence, such that number of page faults is minimized.

        Args:
            sequence (np.ndarray): the sequence that the pyloa.environment.PagingEnvironment instance is going to play
            cache (np.ndarray): the initial cache of the pyloa.environment.PagingEnvironment instance
        """
        cache = np.copy(cache)                                           # cache-copy for simulating self._actions
        time_distance = np.zeros(len(cache), dtype=np.int32)             # temp placeholder for future time_steps
        for time_step, request in enumerate(sequence):                   # for request at time_step
            if request not in cache:                                     # if request not yet in cache
                for index, page in enumerate(cache):                     # for page in cache at index
                    if page == 0 or page not in sequence[time_step+1:]:  # page is init-value or not seen in future
                        self._actions.append(index)                      # -> index is best to evict (flush index)
                        break                                            # magic for-break-else construct (check ref)
                else:                                                    # else case evaluated iff 'break' didn't happen
                    for index, page in enumerate(cache):                 # for page in cache at index
                        # get all future time_steps at which page occurs in future
                        future_time_steps = np.argwhere(sequence[time_step+1:] == page)[0]
                        # if occurence exists, set to first seen time_step, else 'inf'
                        first_occurence = future_time_steps[0] if len(future_time_steps) > 0 else np.iinfo(np.int32).max
                        time_distance[index] = first_occurence
                    # flush index furthest away in the future
                    self._actions.append(np.argmax(time_distance))
                # simulate evict at index
                cache[self._actions[time_step]] = request
            else:                                                        # request is in cache -> do nothing
                self._actions.append(-1)

    # noinspection PyMethodOverriding
    def play_instance(self, environment, episode, writer, invert_cost=True, **kwargs):
        """Agent plays on environment for one episode (on one sequence) by firstly calculating an off-line solution and
        secondly invoking super's play_instance to play it sequentially while 'cheating' with off-line information in
        each function call of act.

        Args:
            environment (pyloa.environment.PagingEnvironment): environment to play on
            episode (int): number depicting how many episodes agent played so far
            writer (tf.summary.FileWriter): tensorboard logging for play sequence
            invert_cost (bool): depicts if costs are inverted after calculation (cost = negative reward)

        Returns:
            tuple: agent will return its calculated metrics of this episode as follows (steps, cost, page_faults)
                    steps (int): depicts the number of actions till the environment terminated
                    cost (int): total cost accumulated during episode
                    page_faults (int): depicts the number of page faults the agent had during episode
        """
        # reset counter, clear last optimal solution
        self._current_time_step = 0
        self._actions.clear()
        # get sequence from environment, calc optimal offline solution on sequence such that self.act is `cheating`
        self._calc_opt(environment.instance, environment.cache)
        # call super-class method on self to emulate steps while tracking metrics
        return PagingAgent.play_instance(self, environment=environment, episode=episode, writer=writer, is_train=False,
                                         invert_cost=invert_cost)

    def act(self, **kwargs):
        """Agent evicts optimal page in cache.

        Returns:
            int: action to be taken
        """
        action = self._actions[self._current_time_step] + 1         # evict page with action 'index+1'
        self._current_time_step += 1                                # increase time step
        return action



#!/usr/bin/env python

"""PyLoa: A research repository for analyzing the performance of classic on-line algorithms vs. modern Machine
Learning, specifically Reinforcement Learning, approaches."""

__author__ = "Matthias Rozanski, Sebastian Schoenitz"
__copyright__ = "Copyright 2019, The PyLoa Project"
__credits__ = []
__license__ = "MIT"
__maintainer__ = "PyLoa Developers"
__email__ = "pyloadevelopers@gmail.com"
__status__ = "Prototype"

import os
import sys
import logging
import argparse
from datetime import datetime
from collections import OrderedDict

sys.path.append(os.path.dirname(sys.path[0]))

from pyloa.utils.load import load_hyperparameters
from pyloa.utils.format import str_iterable
from pyloa.utils.exception import DirectoryNotFoundError, InvalidFileType


def run(args, params, dirs):
    """Invokes the correct run mode.

    Args:
        args (argparse.Namespace): namespace object of parsed arguments via argparse
        params (Dict[str, any]): dictionary defined in config python file
        dirs (OrderedDict[str, str]): dictionary of file/path locations
    """
    if args.runmode == 'train':
        import pyloa._train
        pyloa._train.run(args=args, params=params, dirs=dirs)
    if args.runmode == 'eval':
        import pyloa._eval
        pyloa._eval.run(args=args, params=params, dirs=dirs)
    if args.runmode == 'gen':
        import pyloa._gen
        pyloa._gen.run(args=args, params=params, dirs=dirs)


def main():
    """Generic plumbing: resolve directories, files and paths, parse arguments and set up logging

    Raises:
        DirectoryNotFoundError: if specified config directory does not exist
        FileNotFoundError: if specified config file does not exist
        InvalidFileType: if specified config file is not a python-file
        NotADirectoryError: if specified output directory already exists as a file
    """
    # dirs, files and paths
    _project_directory = os.path.abspath(os.path.realpath(sys.path[0]))     # where python file is located
    _working_directory = os.path.abspath(os.path.realpath(os.getcwd()))     # where user calls python file from
    _config_directory = os.path.join(_working_directory, os.path.join('examples', '0_train_qtable_paging'))
    _config_file = 'hyperparams.py'
    _config_file_path = os.path.join(_config_directory, _config_file)
    _log_directory = _config_directory
    _log_file = 'default.log'
    _log_file_path = os.path.join(_log_directory, _log_file)
    _iso_time = 'E' + datetime.now().replace().isoformat().replace('-', '').replace(':', '').replace('.', '')
    _output_directory = os.path.join(_config_directory, _iso_time)

    # set up arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('runmode', type=str, choices=['train', 'eval', 'gen'],
                        help="depicts the run-mode of _main.py")
    parser.add_argument('-nl', '--no-log',  dest='nlog', default=False, action='store_true',
                        help="deactivates logger (default: '%(default)s')")
    parser.add_argument('-ns', '--no-save', dest='nsave', default=False, action='store_true',
                        help="deactivates saving model (default: '%(default)s')")
    parser.add_argument('-ll', '--loglevel', dest='loglevel', type=str, default='INFO',
                        choices=['CRITICAL', 'ERROR',  'WARNING', 'INFO', 'DEBUG', 'NOTSET'],
                        help="depicts logging level (default: '%(default)s')")
    parser.add_argument('-lf', '--logfile', dest='logfile', type=str, default=_log_file_path,
                        help="destination of log file (default: '%(default)s')")
    parser.add_argument('-lfmt', '--logformat', dest='logformat', type=str,
                        default='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                        help="defines logger message format (default: '%(default)s')")
    parser.add_argument('-ldfmt', '--logdateformat', dest='logdateformat', type=str, default='%m/%d/%Y %I:%M:%S %p',
                        help="defines logger timestamp format (default: '%(default)s')")
    parser.add_argument('-c', '--config',  dest='config', type=str, default=_config_file_path,
                        help="destination of config file (default: '%(default)s')")
    parser.add_argument('-o', '--output-dir', dest='outdir', type=str, default=_output_directory,
                        help="target output directory (default: '%(default)s')")
    args = parser.parse_args()

    # resolve input directories
    _config_directory, _config_file = os.path.split(os.path.expanduser(args.config))
    _config_file_name, _config_file_type = _config_file.rsplit('.', 1) if '.' in _config_file else (_config_file, None)
    _config_directory = os.path.abspath(os.path.realpath(_config_directory))
    # if user specified config directory, place outputs there aswell if not specified otherwise, else use default
    if _config_file_path != args.config and _output_directory == args.outdir:
        _output_directory = os.path.join(_config_directory, _iso_time)
    else:
        _output_directory = os.path.abspath(os.path.realpath(args.outdir))
    _config_file_path = os.path.join(_config_directory, _config_file)
    # if user specified logfile_path, use it, else put logfile in output_dir
    if _log_file_path != args.logfile:
        _log_directory, _log_file = os.path.split(os.path.expanduser(args.logfile))
    else:
        _log_directory, _log_file = _output_directory, os.path.split(os.path.expanduser(args.logfile))[-1]
    _log_directory = os.path.abspath(os.path.realpath(_log_directory))
    _log_file_path = os.path.join(_log_directory, _log_file)

    # check dirs, paths and files for OSErrors
    if not os.path.exists(_config_directory):
        raise DirectoryNotFoundError("config_directory='{}' for config_file='{}' does not exist".format(
                                     _config_directory, _config_file))
    if not os.path.isfile(_config_file_path):
        raise FileNotFoundError("config_file='{}' in config_directory='{}' does not exist or is not a file".format(
                                _config_file, _config_directory))
    if not _config_file_type == 'py':
        raise InvalidFileType("config has to be a '.py' file but found '{}'".format(_config_file_type))
    if not args.nsave or (not args.nlog and _log_directory.startswith(_output_directory)):
        if not os.path.exists(_output_directory):
            os.mkdir(_output_directory)
        elif not os.path.isdir(_output_directory):
            raise NotADirectoryError("can't create directory for output_dir='{}' since file with same path already "
                                     "exists".format(_output_directory))

    # set up logger
    if not args.nlog:
        if not os.path.exists(_log_directory):
            raise DirectoryNotFoundError("log_directory='{}' for log_file='{}' does not exist".format(_log_directory,
                                                                                                      _log_file))
        logging.basicConfig(format=args.logformat, filename=_log_file_path, level=args.loglevel,
                            datefmt=args.logdateformat)
    logging.info("successfully set up logging")
    logging.debug("called {} with following args={}".format(__file__, str_iterable(vars(args))))

    # load params from configuration file
    params = load_hyperparameters(config_file_path=_config_file_path, run_mode=args.runmode)

    # store dirs, paths and files; clean up args before shipping
    dirs = OrderedDict()
    dirs['config_dir'] = _config_directory
    dirs['config_file_path'] = _config_file_path
    dirs['config_file'] = _config_file
    dirs['config_file_name'] = _config_file_name
    dirs['config_file_type'] = _config_file_type
    dirs['log_dir'] = _log_directory
    dirs['log_file_path'] = _log_file_path
    dirs['log_file'] = _log_file
    dirs['project_dir'] = _project_directory
    dirs['working_dir'] = _working_directory
    dirs['output_dir'] = _output_directory
    del args.outdir, args.config, args.logfile, args.loglevel, args.logdateformat, args.logformat
    logging.debug("initialized with the following file/path/directory variables: dirs={}".format(str_iterable(dirs)))
    # start working
    run(args=args, params=params, dirs=dirs)

    # clean up, say fare well
    logging.info("{} terminated".format(__file__))


if __name__ == '__main__':
    main()



from collections.abc import Iterator
from abc import abstractmethod, ABCMeta
import logging


class InstanceGenerator(Iterator, metaclass=ABCMeta):
    """A generator to iteratively generate instance_number-many instances.

    Args:
        instance_number (int): number of instances to create

    Attributes:
        _instance_number (int): number of instances to create
    """
    def __init__(self, instance_number):
        self._instance_number = instance_number
        self._logger = logging.getLogger(__name__)

    def __len__(self):
        """int: returns the total number of instances the generator will generate before depletion"""
        return self._instance_number

    def __iter__(self):
        """creates an generator over self to generate instance

        Returns:
            generator: a generator function to iterate over and generate instances at runtime
        """
        return self.__next__()

    def __next__(self):
        """A generator function that generates len(self)-many instances

        Yields:
            instance: generated instance from a subclass implementation

        Raises:
            StopIteration: raised, when generator is depleted
        """
        for _ in range(len(self)):
            yield self._generate_instance()

    @abstractmethod
    def __str__(self):
        """str: string representation of instance generator"""
        raise NotImplementedError("Must be implemented in subclass.")

    @abstractmethod
    def _generate_instance(self):
        """Create and return a newly generated instance

        Returns:
            instance: returns instance for others to consume

        Raises:
            NotImplementedError: if abstract method is to be invoked
        """
        raise NotImplementedError("Must be implemented in subclass.")





import tensorflow as tf

from .qagent import QAgent
from .vertexagent import VertexAgent


class QLSTMVertexAgent(QAgent, VertexAgent):
    """Agent with a single LSTM-Cell as a Deep Q-Network.

    Args:
        state_dim (int): defines input shape=(_state_dim,) of input states
        action_dim (int): defines the number of different actions the agent can take
        learning_rate (float): learning_rate of agent's model
        units (int): number of units per layer
        discount_factor (float): value at which future Q-target Values are discounted
        memory_size (int): total size of agent's replay memory
        sample_size (int): size of sample randomly drawn from memory to train on per train step
        batch_size (int): size of each batch a sample for training is batched with
        epsilon (float): agent's current chance to exploit his knowledge, else explore action_space
        epsilon_delta (float): increment value of epsilon for each increment call
        epsilon_max (float): maximum value of epsilon
        l2_scale (float): L2-Regularization scale for eval network
        replace_threshold (int): depicts after how many train steps Q-target's variables are replaced with Q-eval's
        save_file (str): location at which agent's checkpoints are stored at

    Attributes:
        _units (int): number of units for LSTM Cell
    """
    def __init__(self,
                 state_dim,
                 action_dim,
                 learning_rate,
                 units,
                 discount_factor,
                 memory_size=5000,
                 sample_size=1000,
                 batch_size=32,
                 epsilon=0.90,
                 epsilon_delta=50e-4,
                 epsilon_max=1.0,
                 l2_scale=0.002,
                 replace_threshold=25,
                 save_file=None,
                 **kwargs):
        super().__init__(
            state_dim=state_dim,
            action_dim=action_dim,
            learning_rate=learning_rate,
            discount_factor=discount_factor,
            memory_size=memory_size,
            sample_size=sample_size,
            batch_size=batch_size,
            epsilon=epsilon,
            epsilon_delta=epsilon_delta,
            epsilon_max=epsilon_max,
            l2_scale=l2_scale,
            replace_threshold=replace_threshold,
            save_file=save_file,
            **kwargs
        )
        # params
        self._units = units
        # init
        self._build_model(**kwargs)
        self._sess.run(tf.group(tf.global_variables_initializer(), tf.local_variables_initializer()))

    # noinspection PyMethodOverriding
    def _build_q_network(self, input_tensor, reuse=tf.AUTO_REUSE, **kwargs):
        """Builds a LSTM Q-Network for Q-Value estimation.

            Args:
                input_tensor (tf.Tensor): accordingly _state or _next_state tensor
                reuse (enum _ReuseMode): tensorflow's reusemode

            Returns:
                tf.Tensor: q_value output of Q-Network
        """
        with tf.variable_scope("vars", reuse=reuse):
            weight = tf.Variable(tf.random_normal([self._units, self._action_dim]), name='weight')
            bias = tf.Variable(tf.random_normal([self._action_dim]), name='bias')
        with tf.variable_scope("cell", reuse=reuse):
            cell = tf.nn.rnn_cell.LSTMCell(
                num_units=self._units,
                forget_bias=1.0,
                state_is_tuple=True,
                activation=tf.nn.leaky_relu,
                dtype=tf.float32,
                reuse=reuse,
            )
        with tf.variable_scope("rnn", reuse=reuse):
            outs, _ = tf.nn.static_rnn(
                cell=cell,
                inputs=[input_tensor],
                dtype=tf.float32
            )
            outs = tf.add(tf.matmul(outs[-1], weight), bias)
        return outs



# define abstract alias imports for other modules
from .agent import Agent
from .rlagent import RLAgent
from .qagent import QAgent
from .pagingagent import PagingAgent
from .vertexagent import VertexAgent
# define alias imports for other modules
from .qmlp import QMLPAgent
from .qlstm import QLSTMAgent
from .qtable import QTableAgent
from .qmlpvertex import QMLPVertexAgent
from .qlstmvertex import QLSTMVertexAgent
from .pagingagent import FIFOAgent, LIFOAgent, BeladysOptimalAgent, LeastFrequentlyUsedAgent, MostFrequentlyUsedAgent, \
    LeastRecentlyUsedAgent, MostRecentlyUsedAgent, RandomAgent
from .vertexagent import FIFOVertexAgent, RandomVertexAgent



import tensorflow as tf
import numpy as np


def tensorboard_scalar(writer, tag, value, step):
    summary = tf.Summary(value=[tf.Summary.Value(tag=tag, simple_value=value)])
    writer.add_summary(summary, step)


def tensorboard_array(writer, tag, value, step):
    # convert to a numpy array
    values = np.asarray(value)
    # create summaries
    _mean = tf.Summary(value=[tf.Summary.Value(tag=tag+'/mean', simple_value=np.mean(values))])
    _min = tf.Summary(value=[tf.Summary.Value(tag=tag+'/min', simple_value=np.min(values))])
    _max = tf.Summary(value=[tf.Summary.Value(tag=tag+'/max', simple_value=np.max(values))])
    _stddev = tf.Summary(value=[tf.Summary.Value(tag=tag+'/stddev', simple_value=np.std(values))])
    # write summaries
    writer.add_summary(_mean, step)
    writer.add_summary(_min, step)
    writer.add_summary(_max, step)
    writer.add_summary(_stddev, step)
    writer.flush()


def tensorboard_text(writer, tag, value, step=0):
    text_tensor = tf.make_tensor_proto(value, dtype=tf.string)
    meta = tf.SummaryMetadata()
    meta.plugin_data.plugin_name = "text"
    summary = tf.Summary()
    summary.value.add(tag=tag, metadata=meta, tensor=text_tensor)
    writer.add_summary(summary)
    writer.flush()


def tensorboard_histo(writer, tag, values, step, bins=1000):
    # convert to a numpy array
    values = np.array(values)
    # create histogram using numpy
    counts, bin_edges = np.histogram(values, bins=bins)
    # fill fields of histogram proto
    hist = tf.HistogramProto()
    hist.min = float(np.min(values))
    hist.max = float(np.max(values))
    hist.num = int(np.prod(values.shape))
    hist.sum = float(np.sum(values))
    hist.sum_squares = float(np.sum(values ** 2))
    # Requires equal number as bins, where the first goes from -DBL_MAX to bin_edges[1]
    # See https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/framework/summary.proto#L30
    # Thus, we drop the start of the first bin
    bin_edges = bin_edges[1:]
    # add bin edges and counts
    for edge in bin_edges:
        hist.bucket_limit.append(edge)
    for c in counts:
        hist.bucket.append(c)
    # create and write summary
    summary = tf.Summary(value=[tf.Summary.Value(tag=tag, histo=hist)])
    writer.add_summary(summary, step)
    writer.flush()
